# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Associate auxiliary data with (almost) any object.

This supports objects that don't allow dynamic attributes, for example,
jax.Array or DynamicJaxprTracer.

Note that aux_data are associated to the exact object and will be cleared after
e.g. copy.copy, jax.tree.map, jnp.astype, etc.
"""

from typing import Any
import weakref


_aux_data: dict[int, tuple[weakref.ref, dict[str, Any]]] = {}
_MISSING = object()


def set(obj: Any, key: str, value: Any):  # pylint: disable=redefined-builtin
  obj_id = id(obj)
  data = _aux_data.get(obj_id, None)
  if data is None or data[0]() is None:
    # _aux_data might be None during shutdown.
    ref = weakref.ref(obj, lambda _: _aux_data and _aux_data.pop(obj_id, None))
    _aux_data[obj_id] = data = (ref, {})
  data[1][key] = value


def get(obj: Any, key: str, default: Any = _MISSING) -> Any:
  data = _aux_data.get(id(obj), None)
  if data is None or data[0]() is None or key not in data[1]:
    if default is _MISSING:
      raise KeyError(f"No auxiliary data {key!r} found for {obj}.")
    return default
  return data[1][key]


def clear(obj: Any):
  _aux_data.pop(id(obj), None)



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Post-training quantization (PTQ)."""

import functools
from typing import Any, Callable, Generic, Sequence, TypeVar

from flax import linen as nn
from flax import nnx
import flax.linen.dtypes
import jax
from qwix._src import averaging
from qwix._src import flax_util
from qwix._src import qconfig
from qwix._src.core import conv_general
from qwix._src.core import dot
from qwix._src.core import dot_general
from qwix._src.core import einsum
from qwix._src.core import qarray


ArrayTypeVar = TypeVar('ArrayTypeVar', jax.Array, qarray.QArray)


@flax.struct.dataclass
class WithAux(Generic[ArrayTypeVar]):
  """An array/QArray with auxiliary information.

  The main purpose of this class is to embed the how to quantize information
  into the param tree, such that the quantize_params() function can quantize
  params without knowing the model structure.

  Attributes:
    array: The underlying array.
    how: How the array is quantized, which is used by quantize_params so that it
      knows how to quantize the original weights.
    value: Satisfies the nnx.Variable interface.
  """

  array: ArrayTypeVar
  how: qarray.HowToQuantize = flax.struct.field(pytree_node=False)

  # This allows us to appear like nnx.Variable.
  value = property(flax_util.unbox)
  shape = property(lambda self: flax_util.unbox(self.array).shape)
  ndim = property(lambda self: flax_util.unbox(self.array).ndim)
  __getitem__ = lambda self, key: jax.tree.map(lambda x: x[key], self.value)

  def reshape(self, *shape):
    if len(shape) == 1:
      try:
        shape = tuple(shape[0])
      except TypeError:
        pass
    if tuple(self.shape) != tuple(shape):
      raise ValueError(
          'PTQ weights should already have the target shape. Got'
          f' {self.shape=} but {shape=} is requested.'
      )
    return self


class PtqProvider(qconfig.QuantizationProvider):
  """Quantization provider for PTQ.

  In PTQ mode, weights needs to be pre-quantized. However, Qwix doesn't know
  about how to quantize them until the actual ops get called. To solve this,
  we still initialize the original weights when the model is initialized, but
  we replace them with the quantized weights when the ops are called.

  * It should be invisible to users in Flax linen because `module.init` will
    call both the setup() and __call__() methods.
  * If memory usage is a concern, wrapping `module.init` with jit or eval_shape
    should avoid materializing the original weights.
  * NNX can use the same trick so we don't need to intercept nnx.Param.
  * This approach allows the original weights to be supplied during `apply`,
    and will actually quantize them correctly. This can be an alternative to
    `quantize_params` if partial param quantization is not needed.
  """

  def __init__(
      self,
      rules: Sequence[qconfig.QuantizationRule],
      *,
      _qarray_module=qarray,
      _dot_general_fn=dot_general.dot_general,
      _einsum_fn=einsum.einsum,
      _conv_general_dilated_fn=conv_general.conv_general_dilated,
  ):
    """Initializes the PTQ provider."""
    super().__init__(rules)
    self._qarray_module = _qarray_module
    self._dot_general_fn = _dot_general_fn
    self._einsum_fn = _einsum_fn
    self._conv_general_dilated_fn = _conv_general_dilated_fn

  def dot_general(
      self,
      lhs: jax.Array,
      rhs: jax.Array | WithAux[qarray.QArray],
      dimension_numbers: jax.lax.DotDimensionNumbers,
      precision: jax.lax.PrecisionLike = None,
      preferred_element_type: jax.typing.DTypeLike | None = None,
      *,
      out_sharding: jax.sharding.NamedSharding | None = None,
  ) -> jax.Array:
    rule, op_id = self._get_current_rule_and_op_id('dot_general')
    if rule is None or rule.weight_qtype is None:
      return jax.lax.dot_general(
          lhs,
          rhs,
          dimension_numbers,
          precision=precision,
          preferred_element_type=preferred_element_type,
          out_sharding=out_sharding,
      )

    get_how_to_quantize = functools.partial(
        dot_general.get_how_to_quantize,
        dimension_numbers=dimension_numbers,
        ndims=(len(lhs.shape), len(rhs.shape)),
        tile_size=rule.tile_size,
    )

    # Prepare rhs.
    if isinstance(rhs, WithAux):  # weight, already quantized
      rhs = rhs.array
    elif weight_name := flax_util.find_param(rhs):  # weight, not quantized
      rhs_how = get_how_to_quantize(
          for_lhs=False,
          qtype=rule.weight_qtype,
          calibration_method=rule.weight_calibration_method,
      )
      rhs = create_quantized_param(
          weight_name, rhs, rhs_how, _qarray_module=self._qarray_module
      ).array
    elif rule.act_qtype is not None:  # act
      rhs_how = get_how_to_quantize(
          for_lhs=False,
          qtype=rule.act_qtype,
          calibration_method=rule.act_calibration_method,
      )
      rhs = quantize_act(
          rhs, rhs_how, rule, op_id + '_rhs', _qarray_module=self._qarray_module
      )

    # Prepare lhs.
    if rule.act_qtype is not None:
      lhs_how = get_how_to_quantize(
          for_lhs=True,
          qtype=rule.act_qtype,
          calibration_method=rule.act_calibration_method,
      )
      lhs = quantize_act(
          lhs, lhs_how, rule, op_id + '_lhs', _qarray_module=self._qarray_module
      )
    return self._dot_general_fn(
        lhs, rhs, dimension_numbers, out_sharding=out_sharding
    )

  def einsum(
      self,
      einsum_str: str,
      *operands: jax.Array,
      precision: jax.lax.PrecisionLike = None,
      preferred_element_type: jax.typing.DTypeLike | None = None,
      _dot_general: Callable[..., jax.Array] = jax.lax.dot_general,  # pylint: disable=invalid-name
      out_sharding=None,
  ) -> jax.Array:
    rule, op_id = self._get_current_rule_and_op_id('einsum')
    if rule is None or rule.weight_qtype is None:
      return jax.numpy.einsum(
          einsum_str,
          *operands,
          precision=precision,
          preferred_element_type=preferred_element_type,
          _dot_general=_dot_general,
          out_sharding=out_sharding,
      )
    if not isinstance(einsum_str, str) or len(operands) != 2:
      raise ValueError(f'Unsupported einsum format: {einsum_str=} {operands=}')

    lhs, rhs = operands
    get_how_to_quantize = functools.partial(
        einsum.get_how_to_quantize,
        einsum_str=einsum_str,
        ndims=(len(lhs.shape), len(rhs.shape)),
        tile_size=rule.tile_size,
    )

    # Prepare rhs.
    if isinstance(rhs, WithAux):  # weight, already quantized
      rhs = rhs.array
    elif weight_name := flax_util.find_param(rhs):  # weight, not quantized
      rhs_how = get_how_to_quantize(
          for_lhs=False,
          qtype=rule.weight_qtype,
          calibration_method=rule.weight_calibration_method,
      )
      rhs = create_quantized_param(
          weight_name, rhs, rhs_how, _qarray_module=self._qarray_module
      ).array
    elif rule.act_qtype is not None:  # act
      rhs_how = get_how_to_quantize(
          for_lhs=False,
          qtype=rule.act_qtype,
          calibration_method=rule.act_calibration_method,
      )
      rhs = quantize_act(
          rhs, rhs_how, rule, op_id + '_rhs', _qarray_module=self._qarray_module
      )

    # Prepare lhs.
    if rule.act_qtype is not None:
      lhs_how = get_how_to_quantize(
          for_lhs=True,
          qtype=rule.act_qtype,
          calibration_method=rule.act_calibration_method,
      )
      lhs = quantize_act(
          lhs, lhs_how, rule, op_id + '_lhs', _qarray_module=self._qarray_module
      )
    return self._einsum_fn(einsum_str, lhs, rhs)

  def conv_general_dilated(
      self,
      lhs: jax.Array,
      rhs: jax.Array | WithAux[qarray.QArray],
      window_strides: Sequence[int],
      padding: str | Sequence[tuple[int, int]],
      lhs_dilation: Sequence[int] | None = None,
      rhs_dilation: Sequence[int] | None = None,
      dimension_numbers: jax.lax.ConvGeneralDilatedDimensionNumbers = None,
      feature_group_count: int = 1,
      batch_group_count: int = 1,
      precision: jax.lax.PrecisionLike = None,
      preferred_element_type: jax.typing.DTypeLike | None = None,
  ) -> jax.Array:
    rule, op_id = self._get_current_rule_and_op_id('conv_general_dilated')
    if rule is None or rule.weight_qtype is None:
      return jax.lax.conv_general_dilated(
          lhs,
          rhs,
          window_strides,
          padding,
          lhs_dilation=lhs_dilation,
          rhs_dilation=rhs_dilation,
          dimension_numbers=dimension_numbers,
          feature_group_count=feature_group_count,
          batch_group_count=batch_group_count,
          precision=precision,
          preferred_element_type=preferred_element_type,
      )
    dimension_numbers = jax.lax.conv_dimension_numbers(
        lhs.shape, rhs.shape, dimension_numbers
    )

    # Prepare rhs.
    if isinstance(rhs, WithAux):  # weight, already quantized
      rhs = rhs.array
    else:
      weight_name = flax_util.find_param(rhs)
      rhs_how = conv_general.get_how_to_quantize(
          dimension_numbers=dimension_numbers,
          for_lhs=False,
          qtype=rule.weight_qtype,
          calibration_method=rule.weight_calibration_method,
      )
      rhs = create_quantized_param(
          weight_name, rhs, rhs_how, _qarray_module=self._qarray_module
      ).array

    # Prepare lhs.
    if rule.act_qtype != rule.weight_qtype:
      raise ValueError(
          'conv_general_dilated requires same act_qtype and weight_qtype. Got:'
          f' {rule.act_qtype=} {rule.weight_qtype=}'
      )
    lhs_how = conv_general.get_how_to_quantize(
        dimension_numbers=dimension_numbers,
        for_lhs=True,
        qtype=rule.act_qtype,
        calibration_method=rule.act_calibration_method,
    )
    lhs = quantize_act(
        lhs, lhs_how, rule, op_id + '_lhs', _qarray_module=self._qarray_module
    )
    return self._conv_general_dilated_fn(
        lhs,
        rhs,
        window_strides,
        padding,
        lhs_dilation=lhs_dilation,
        rhs_dilation=rhs_dilation,
        dimension_numbers=dimension_numbers,
        feature_group_count=feature_group_count,
        batch_group_count=batch_group_count,
    )

  def nn_param(self, module: nn.Module, name: str, *args, **kwargs):
    """Intercepts nn.Module.param to handle quantized params."""
    # Don't check the shape if the param is already quantized.
    existing_param = module.get_variable('params', name)
    if isinstance(existing_param, WithAux):
      return nn.unbox(existing_param)
    return module.param(name, *args, **kwargs)

  def promote_dtype(self, *args, **kwargs):
    """Intercepts flax.{linen,nnx.nn}.dtypes.promote_dtype to handle quantized params."""
    if len(args) == 1 and isinstance(args[0], Sequence):
      args = args[0]  # nnx version
    # Skip WithAux.
    array_args = [x if isinstance(x, jax.Array) else None for x in args]
    array_args = flax.linen.dtypes.promote_dtype(*array_args, **kwargs)
    return [x if x is not None else y for x, y in zip(array_args, args)]

  def dot(
      self,
      a: jax.Array,
      b: jax.Array | WithAux[qarray.QArray],
      precision: jax.lax.PrecisionLike = None,
      preferred_element_type: jax.typing.DTypeLike | None = None,
      out_sharding=None,
  ):
    """Intercepts jax.numpy.dot."""
    return dot.dot(
        a,
        b,
        precision=precision,
        preferred_element_type=preferred_element_type,
        out_sharding=out_sharding,
        _qwix_dot_general=self.dot_general,
    )

  def get_intercept_map(self):
    """Used for interception."""
    return super().get_intercept_map() | {
        'jax.lax.conv_general_dilated': self.conv_general_dilated,
        'jax.lax.dot_general': self.dot_general,
        'jax.numpy.dot': self.dot,
        'jax.numpy.einsum': self.einsum,
        'flax.linen.Module.param': self.nn_param,
        'flax.linen.dtypes.promote_dtype': self.promote_dtype,
        'flax.nnx.nn.dtypes.promote_dtype': self.promote_dtype,
    }


def quantize_act(
    array: jax.Array,
    how: qarray.HowToQuantize,
    rule: qconfig.QuantizationRule,
    act_name: str,
    *,
    _qarray_module=qarray,
) -> qarray.QArray:
  """Quantizes the input activation with support for static scale."""
  if not rule.act_static_scale:
    return _qarray_module.quantize(array, how)

  # Construct the scale and zero_point from the quant stats, if available.
  # This is useful in NNX when a PTQ model is converted from a QAT model.
  # We delete the quant_stat after the first forward pass so that the PTQ
  # model appears the same as a regular one.
  quant_stat = flax_util.get_and_delete_variable('quant_stats', act_name)

  def init():
    if quant_stat is not None:
      aggregator = averaging.SimpleMovingAverage()
      calibration = aggregator.get_calibration(quant_stat)
    else:
      calibration = _qarray_module.calibrate(array, how)
      # Apply act_batch_axes for static scale.
      calibration = jax.tree.map(
          lambda x: x.mean(axis=rule.act_batch_axes, keepdims=True), calibration
      )
    nonlocal zp
    scale, zp = _qarray_module.compute_scale_zero_point(calibration, how.qtype)
    # Wrap scale in WithAux because quantize_params needs to know the qtype.
    return WithAux(scale, how)

  zp = None
  scale = flax_util.get_or_create_param(act_name + '_scale', init)
  if zp is not None:
    zp = flax_util.get_or_create_param(act_name + '_zero_point', lambda: zp)
  return _qarray_module.quantize_with_scale_zero_point(
      array, how.qtype, scale.array, zp
  )


def create_quantized_param(
    name: str,
    value: jax.Array,
    how: qarray.HowToQuantize,
    *,
    _qarray_module=qarray,
) -> WithAux[qarray.QArray]:
  """Creates the quantized param and replaces the original param in the module.

  Args:
    name: The name of the param in the module.
    value: The unquantized jax.Array.
    how: How to quantize the param.
    _qarray_module: The qarray module to use. Useful for extending.

  Returns:
    An unboxed WithAux.
  """
  unboxed = WithAux(_qarray_module.quantize(value, how), how)

  # The following code is about replacing the saved param with WithAux, with
  # correct metadata.

  module = flax_util.get_current_module()
  if isinstance(module, nn.Module):
    if not module.is_initializing():
      raise ValueError(
          "It seems you're feeding an unquantized param to a quantized model."
      )
    param = module.get_variable('params', name)
    boxed = jax.tree.map(
        lambda value: flax_util.update_boxed(param, value=value), unboxed
    )
    module.put_variable('params', name, boxed)
  elif isinstance(module, nnx.Module):
    param = getattr(module, name)
    boxed = jax.tree.map(
        lambda value: flax_util.update_boxed(param, value=value), unboxed
    )
    setattr(module, name, boxed)

  return unboxed


def quantize_params(
    params: Any,
    abstract_quantized_params: Any,
    quant_stats: Any = flax.core.FrozenDict(),
    *,
    _qarray_module=qarray,
    allow_extra_params: bool = False,
) -> Any:
  """Quantize the param tree for PTQ.

  This function provides advanced param quantization for PTQ. It is useful
  when the original params are too large to fit in the HBM.

  Args:
    params: The floating-point param tree to quantize, which is usually
      generated by the original or QAT model. The tree doesn't need to be
      complete and can be a subtree of the whole param tree. In NN, the tree
      needs to be unboxed, i.e. nn.unbox(). In NNX, the tree needs to be a pure
      dict, i.e. nnx.to_pure_dict().
    abstract_quantized_params: The param tree generated by the PTQ model, which
      can be abstract with jax.ShapeDtypeStruct as leaves instead of jax.Array.
      In NN, the tree may contain AxisMetadata. In NNX, this should be the PTQ
      model itself, possibly abstract.
    quant_stats: The quantization statistics. This is only used in SRQ.
      _qarray_module: The qarray module to use. Useful for extending.
    allow_extra_params: If True, allow the params to contain extra parameters
      that are not present in the abstract_quantized_params, e.g., loss
      computation that are not triggered in PTQ.

  Returns:
    The quantized param tree, which has the same structure as the input params
    but with quantized leaves.
  """

  def get_value_from_path(obj, path: tuple[str, ...]):
    for key in path:
      if obj is None:
        return None
      obj = obj.get(key) if isinstance(obj, dict) else getattr(obj, key)
    return obj

  quantized_params = {}
  for path, param in flax.traverse_util.flatten_dict(params).items():
    if not isinstance(param, jax.Array):
      raise TypeError(f'params is not a pure dict of jax.Array: {type(param)}')
    abs_param = get_value_from_path(abstract_quantized_params, path)
    if abs_param is None:
      if allow_extra_params:
        continue
      else:
        raise ValueError(
            f'Quantized param {path} is not found in the abstract quantized'
            ' params.'
        )
    if isinstance(abs_param, WithAux):
      # The param might not be in the shape needed for compute, in case the
      # module reshapes before compute. Abstract param has the compute shape.
      param = param.reshape(abs_param.shape)
      param = abs_param.replace(
          array=_qarray_module.quantize(param, abs_param.how)
      )
    quantized_params[path] = param

  # SRQ only: compute scale and zero_point from the quant_stats.
  all_quant_stats_paths = {
      path[:-1] for path in flax.traverse_util.flatten_dict(quant_stats)
  }
  for path in all_quant_stats_paths:
    quant_stat = get_value_from_path(quant_stats, path)
    if quant_stat['count'] == 0:
      raise ValueError(f'quant_stats is not initialized for {path}.')

    # Get the act_qtype from the scale, which is a WithAux[jax.Array].
    scale_path = (*path[:-1], path[-1] + '_scale')
    abs_scale = get_value_from_path(abstract_quantized_params, scale_path)
    assert isinstance(abs_scale, WithAux)
    act_qtype = abs_scale.how.qtype

    calibration = averaging.SimpleMovingAverage().get_calibration(quant_stat)
    scale, zero_point = _qarray_module.compute_scale_zero_point(
        calibration, act_qtype
    )
    quantized_params[scale_path] = abs_scale.replace(array=scale)
    if zero_point is not None:
      quantized_params[(*path[:-1], path[-1] + '_zero_point')] = zero_point

  if isinstance(abstract_quantized_params, nnx.Module):
    # Convert WithAux to a pure dict so that nnx.update() can work.
    quantized_params = nnx.to_pure_dict(nnx.state(quantized_params))

  return flax.traverse_util.unflatten_dict(quantized_params)



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Quantized Array."""

import dataclasses
import functools
from typing import Callable, Collection, Mapping, Sequence, TypeAlias
import flax.struct
import jax
from jax import numpy as jnp
from qwix._src.core import numerics


# ---------------------------------------------
# QArray definition and common functions.
# ---------------------------------------------


@flax.struct.dataclass
class QArray:
  """A quantized array implementation with subchannel support.

  The following conditions hold:
    * qvalue.shape == original.shape
    * len(scale.shape) == len(original.shape)
    * len(scale.shape) == len(zero_point.shape)
    * To enable subchannel quantization, scale and zero_point can be
      "generic broadcasted" to original.shape, which means
        all(o % s == 0 for o, s in zip(original.shape, scale.shape))
    * original ≈ (qvalue - zero_point) * generic_broadcast(
        scale, original.shape)

  Attributes:
    qvalue: The quantized value.
    scale: The scale used to quantize the value.
    zero_point: The quantization value that represents the exact floating-point
      value 0, or None if in symmetric quantization.
    qtype: The logical type of the qvalue, which could be different from the
      dtype used for storage in qvalue. If None, the qvalue's dtype will be used
      as the logical type.
  """

  qvalue: jax.Array
  scale: jax.Array
  zero_point: jax.Array | None = None
  qtype: jax.typing.DTypeLike = flax.struct.field(
      pytree_node=False, default=None
  )

  # Array-like methods.
  shape = property(lambda self: self.qvalue.shape)
  ndim = property(lambda self: self.qvalue.ndim)
  dtype = property(lambda self: self.scale.dtype)
  T = property(lambda self: self.transpose())
  mT = property(lambda self: jax.tree.map(lambda x: x.mT, self))  # pylint: disable=invalid-name

  @property
  def scale_tile_shape(self) -> tuple[int, ...]:
    """Returns the tile shape for the scale values."""
    return tuple(o // s for o, s in zip(self.shape, self.scale.shape))

  @property
  def zero_point_tile_shape(self) -> tuple[int, ...] | None:
    """Returns the tile shape for the zero point values."""
    if self.zero_point is None:
      return None
    return tuple(o // s for o, s in zip(self.shape, self.zero_point.shape))

  def reshape(self, *new_shape) -> 'QArray':
    return reshape(self, *new_shape)

  def transpose(self, *args) -> 'QArray':
    return jax.tree.map(lambda x: x.transpose(*args), self)

  def __getitem__(self, idx) -> 'QArray':
    return rewriting_take(self, idx)

  def __post_init__(self):
    if self.qtype is None:
      object.__setattr__(self, 'qtype', self.qvalue.dtype)

  def astype(self, dtype: jax.typing.DTypeLike) -> 'QArray':
    """Cast the dequant type to the given dtype."""
    return dataclasses.replace(self, scale=self.scale.astype(dtype))

  def swapaxes(self, axis1: int, axis2: int) -> 'QArray':
    return jax.tree.map(lambda x: x.swapaxes(axis1, axis2), self)


def reshape(array: QArray, *new_shape) -> QArray:
  """Reshapes the array, which is not always feasible."""
  if len(new_shape) == 1:
    try:
      new_shape = tuple(new_shape[0])
    except TypeError:
      pass

  prod = lambda s: functools.reduce(lambda x, y: x * y, s, 1)

  old_shape = array.shape
  if prod(old_shape) != prod(new_shape):
    raise ValueError(f'Cannot reshape {old_shape} into {new_shape}.')

  # Group the old shape and the new shape. e.g. for (2, 2, 6) -> (4, 1, 3, 2),
  # the groups are (2, 2) -> (4,) and (6,) -> (3, 2) with 1 ignored.
  groups = []
  last_group_old, last_group_new = [], []
  i, j = 0, 0
  while i < len(old_shape) or j < len(new_shape):
    # INVARIANT: if prod(last_group_old) == prod(last_group_new), they
    # must be both empty.
    if prod(last_group_old) < prod(last_group_new):
      last_group_old.append(old_shape[i])
      i += 1
    elif prod(last_group_old) > prod(last_group_new):
      last_group_new.append(new_shape[j])
      j += 1
    elif i < len(old_shape) and (
        j >= len(new_shape) or old_shape[i] <= new_shape[j]
    ):
      last_group_old.append(old_shape[i])
      i += 1
    elif j < len(new_shape):
      last_group_new.append(new_shape[j])
      j += 1
    if prod(last_group_old) == prod(last_group_new):
      groups.append((tuple(last_group_old), tuple(last_group_new)))
      last_group_old, last_group_new = [], []
  assert not last_group_old and not last_group_new

  def reshape_by_groups(x: jax.Array) -> jax.Array:
    i = 0
    target_shape = []
    for old, new in groups:
      actual_shape = x.shape[i : i + len(old)]
      actual_size = prod(actual_shape)
      if actual_shape == old:
        target_shape.extend(new)
      elif actual_size == 1:
        target_shape.extend(1 for _ in new)
      elif actual_size == actual_shape[0] and new[0] % actual_shape[0] == 0:
        # Channelwise-preserving reshape, e.g.
        #   qvalue: (2, 2, 6) -> (4, 3, 2)
        #   scale:  (2, 1, 3) -> (2, 3, 1)  OK
        #   scale:  (1, 2, 2) -> NOT OK
        target_shape.extend([actual_shape[0]] + [1] * (len(new) - 1))
      else:
        raise ValueError(
            f'Cannot reshape {old_shape} into {new_shape} for {x.shape}.'
        )
      i += len(old)
    return x.reshape(target_shape)

  return jax.tree.map(reshape_by_groups, array)


def rewriting_take(array: QArray, idx) -> QArray:
  """Returns array[*idx]."""
  idx = list(idx) if isinstance(idx, tuple | list) else [idx]

  actual_len = sum(i is not None and i is not Ellipsis for i in idx)
  if Ellipsis in idx:
    assert idx.count(Ellipsis) == 1
    index = idx.index(Ellipsis)
    idx[index : index + 1] = [slice(None)] * (array.ndim - actual_len)
  else:
    idx += [slice(None)] * (array.ndim - actual_len)

  def take(x: jax.Array) -> jax.Array:
    actual_idx = []
    i = 0
    for a, b in zip(array.shape, x.shape):
      while idx[i] is None:
        actual_idx.append(None)
        i += 1
      if a == b or idx[i] == slice(None):
        actual_idx.append(idx[i])
      elif b == 1:
        if isinstance(idx[i], int) or (
            isinstance(idx[i], jax.Array) and idx[i].ndim == 0
        ):
          actual_idx.append(0)
        else:
          actual_idx.append(slice(None))
      elif isinstance(idx[i], int):
        actual_idx.append(idx[i] // (a // b))
      else:
        raise ValueError(f'Unsupported indexing {idx} for {array.shape}.')
      i += 1
    actual_idx.extend(idx[i:])
    return x[tuple(actual_idx)]

  return jax.tree.map(take, array)


def validate_qarray(array: QArray):
  """Validates the internal consistency of a QArray."""
  if not isinstance(array.qvalue, jax.Array):
    return  # don't check if qvalue is nn.Partitioned, pl.BlockSpec, etc.
  if array.qvalue.ndim != array.scale.ndim:
    raise ValueError(
        f'Scale {array.scale.shape} should have the same rank as qvalue'
        f' {array.qvalue.shape}.'
    )
  if not all(a % b == 0 for a, b in zip(array.qvalue.shape, array.scale.shape)):
    raise ValueError(
        f'Scale {array.scale.shape} should be broadcastable to qvalue'
        f' {array.qvalue.shape}.'
    )
  if array.qvalue.dtype.itemsize > 1:
    raise ValueError(f'{array.qvalue.dtype} is not a valid type for qvalue.')
  if array.scale.dtype not in (jnp.bfloat16, jnp.float32, jnp.float64):
    raise ValueError(f'{array.scale.dtype} is not a valid type for scale.')
  if array.zero_point is not None:
    if array.zero_point.ndim != array.qvalue.ndim:
      raise ValueError(
          f'Zero point {array.zero_point.shape} should have the same rank as'
          f' qvalue {array.qvalue.shape}.'
      )
    if not all(
        a % b == 0 for a, b in zip(array.qvalue.shape, array.zero_point.shape)
    ):
      raise ValueError(
          f'Zero point {array.zero_point.shape} should be broadcastable to'
          f' qvalue {array.qvalue.shape}.'
      )
    if array.zero_point.dtype != array.qvalue.dtype:
      raise ValueError(
          f'Zero point {array.zero_point.dtype} should have the same dtype as'
          f' qvalue {array.qvalue.dtype}.'
      )


# ---------------------------------------------
# Quantization and dequantization of QArray.
# ---------------------------------------------


@dataclasses.dataclass(slots=True, frozen=True, kw_only=True)
class HowToQuantize:
  """Determines how to quantize an array."""

  # The the logical type of the qvalue.
  # E.g. jnp.int8, jnp.int4, jnp.float8_*, nf4, etc.
  # Actual qvalue dtype is determined by the quantization method.
  qtype: jax.typing.DTypeLike
  # Channelwise axes will have individual scales, which has the same effect
  # as setting their tile sizes to 1 in tiled_axes.
  channelwise_axes: Collection[int] = ()
  # Tiled axes have subchannel quantization enabled. The value is a mapping
  # from the tiled axis to the tile size. If the tile size is a float, it has
  # to be "1 / tile_count" and the actual tile size will be
  # round(axis_size * tile_size). Note that 1 and 1.0 have very different
  # meanings: a tile size of 1 means to use per-channel scale, while a
  # tile size of 1.0 means to use shared scale.
  tiled_axes: Mapping[int, int | float] = dataclasses.field(
      default_factory=dict
  )
  # The calibration method to use. The format is <method>[,<args>], e.g.
  # "absmax" or "fixed,-10,10". Check calibrate() for supported methods.
  calibration_method: str = 'absmax'
  # Noise function to use for stochastic rounding.
  noise_fn: numerics.NoiseFn | None = None


ShapeT: TypeAlias = Sequence[int]
MaybeQArray: TypeAlias = jax.Array | QArray


def get_scale_shape(array_shape: ShapeT, how: HowToQuantize) -> ShapeT:
  """Returns the scale shape."""
  if set(how.channelwise_axes) & how.tiled_axes.keys():
    raise ValueError('The same axis cannot be both channelwise and tiled.')
  scale_shape = []
  for axis, dim in enumerate(array_shape):
    if axis in how.channelwise_axes:
      scale_shape.append(dim)
    elif axis in how.tiled_axes:
      tile_size = how.tiled_axes[axis]
      if isinstance(tile_size, float):
        tile_size = round(dim * tile_size)
      if tile_size <= 0 or dim % tile_size != 0:
        raise ValueError(f'{array_shape} cannot be tiled as {how.tiled_axes}.')
      scale_shape.append(dim // tile_size)
    else:
      scale_shape.append(1)
  return tuple(scale_shape)


def transpose_array(
    array: jax.Array, transpose: Sequence[int | None]
) -> jax.Array:
  """Enhanced version of jnp.transpose.

  * It allows missing and new axes in the transpose list. The missing axes will
    be squeezed away, and the new axes will be added with size 1.
  * It's pallas-friendly as it will try to use reshape and simpler transpose
    instead of direct transpose when possible.

  Args:
    array: The array to transpose.
    transpose: The transpose order.

  Returns:
    The transposed array.
  """
  if any(l > 1 for a, l in enumerate(array.shape) if a not in transpose):
    raise ValueError(f'Cannot transpose {array.shape} as {transpose}.')
  used_axes = [a for a in transpose if a is not None and array.shape[a] != 1]
  # If used_axes is already in order, no actual transpose is needed and we can
  # just reshape the array.
  if sorted(used_axes) == used_axes:
    return array.reshape(
        [1 if a is None else array.shape[a] for a in transpose]
    )
  # Because transpose is not generally supported in pallas kernels, we try to
  # avoid complex transposes here by calling squeeze first. For example, for
  # transpose [1, 2, None], instead of just calling transpose(1, 2, 0), we call
  # squeeze(0).expand(2).
  return (
      array.squeeze([a for a in range(array.ndim) if a not in used_axes])
      .transpose([sum(i < a for i in used_axes) for a in used_axes])
      .reshape([1 if a is None else array.shape[a] for a in transpose])
  )


def split_axis(
    array: jax.Array, tiled_axes: Mapping[int, int | float]
) -> jax.Array:
  """Reshape the array where the axis is split into (tile_count, tile_size)."""
  new_shape = []
  for axis, dim in enumerate(array.shape):
    if axis in tiled_axes:
      tile_size = tiled_axes[axis]
      if isinstance(tile_size, float):
        tile_size = round(dim * tile_size)
      if dim % tile_size != 0:
        raise ValueError(f'{array.shape} cannot be tiled as {tiled_axes}.')
      new_shape.append(dim // tile_size)
      new_shape.append(tile_size)
    else:
      new_shape.append(dim)
  return array.reshape(new_shape)


def get_tiled_axes(array: QArray) -> dict[int, int]:
  """Infers the tiled axes from a QArray.

  Args:
    array: The QArray to infer the tiled axes from.

  Returns:
    A dict from tiled axis to tile size.
  """
  tiled_axes = {}
  for i, (j, k) in enumerate(zip(array.qvalue.shape, array.scale.shape)):
    if j != k and k != 1:
      tiled_axes[i] = j // k
  return tiled_axes


def call_with_generic_broadcast(
    op: Callable[[jax.Array, jax.Array], jax.Array], x: jax.Array, y: jax.Array
):
  """Call an element-wise binary op with generic broadcast."""
  assert x.ndim == y.ndim
  x_shape, y_shape, o_shape = [], [], []
  for a, b in zip(x.shape, y.shape):
    o_shape.append(max(a, b))
    if a == b or a == 1 or b == 1:
      x_shape.append(a)
      y_shape.append(b)
    elif a % b == 0:
      x_shape.extend((b, a // b))
      y_shape.extend((b, 1))
    elif b % a == 0:
      x_shape.extend((a, 1))
      y_shape.extend((a, b // a))
    else:
      raise ValueError(f'Cannot broadcast between {x.shape} {y.shape}')
  return op(x.reshape(x_shape), y.reshape(y_shape)).reshape(o_shape)


def calibrate(array: jax.Array, how: HowToQuantize) -> dict[str, jax.Array]:
  """Calibrates the array.

  Args:
    array: The array to calibrate.
    how: How to quantize the array.

  Returns:
    A dict of quantization statistics, e.g. {'min': ..., 'max': ...} for
    asymmetric quantization, or {'absmax': ...} for symmetric quantization.
    Each value in the dict has the same shape as the (expected) scale.
  """
  if how.qtype == 'mxfp8' or how.qtype == 'mxfp4':
    last_axis = array.ndim - 1
    how = dataclasses.replace(
        how,
        channelwise_axes=list(range(last_axis)),
        tiled_axes={last_axis: 32},
    )
  reduce_axes = []  # axes to calibrate.
  tiled_axes_offset = 0
  for axis, _ in enumerate(array.shape):
    if axis in how.channelwise_axes:
      continue  # no reduce needed.
    if axis in how.tiled_axes:
      tiled_axes_offset += 1  # reduce the tile_size rather than num_tiles.
    reduce_axes.append(axis + tiled_axes_offset)

  # The returned calibration values should have the same shape as the scale.
  shape = get_scale_shape(array.shape, how)
  array = split_axis(array, how.tiled_axes)

  # Parse the calibration method.
  method, *args = how.calibration_method.lower().split(',')
  args = [float(a) for a in args]
  if method == 'minmax':
    min_array = jnp.min(array, axis=reduce_axes, keepdims=True)
    max_array = jnp.max(array, axis=reduce_axes, keepdims=True)
    # Ensure min_array <= 0 <= max_array so that 0 can be accurately quantized.
    min_array = jnp.clip(min_array, max=0)
    max_array = jnp.clip(max_array, min=0)
    if args:  # args[0] is the scale factor.
      min_array = min_array * args[0]
      max_array = max_array * args[0]
    return {'min': min_array.reshape(shape), 'max': max_array.reshape(shape)}
  elif method == 'absmax':
    absmax = jnp.max(jnp.abs(array), axis=reduce_axes, keepdims=True)
    if args:  # args[0] is the scale factor.
      absmax = absmax * args[0]
    return {'absmax': absmax.reshape(shape)}
  elif method == 'rms':
    rms = jnp.sqrt(jnp.mean(jnp.square(array), axis=reduce_axes, keepdims=True))
    if not args:
      raise ValueError('A scale factor is required for RMS calibration.')
    return {'absmax': (rms * args[0]).reshape(shape)}
  elif method == 'fixed':
    if len(args) not in (1, 2):
      raise ValueError('A fixed range is required for fixed calibration.')
    if len(args) == 1:
      args = (-args[0], args[0])
    if args[0] > 0 or args[1] < 0 or args[0] >= args[1]:
      raise ValueError('The range must contain 0 and be non-empty.')
    # Fixed calibration is always per-tensor.
    shape = tuple(1 for _ in shape)
    if args[0] + args[1] == 0:
      return {'absmax': jnp.full(shape, args[1], array.dtype)}
    return {
        'min': jnp.full(shape, args[0], array.dtype),
        'max': jnp.full(shape, args[1], array.dtype),
    }
  else:
    raise ValueError(f'Unsupported calibration: {how.calibration_method}')


def compute_scale_zero_point(
    calibration: Mapping[str, jax.Array], qtype: jax.typing.DTypeLike
) -> tuple[jax.Array, jax.Array | None]:
  """Computes the scale and zero_point from the calibration result.

  Args:
    calibration: The calibration returned by calibrate().
    qtype: The dtype of the qvalue.

  Returns:
    A tuple of the scale and zero_point. The zero_point is None in symmetric
    quantization.
  """
  if 'min' in calibration and 'max' in calibration:
    qmin, qmax = numerics.get_asymmetric_bound(qtype)
    scale = (calibration['max'] - calibration['min']) / (qmax - qmin)
    scale = jnp.where(scale == 0, 1, scale)  # Scale shouldn't be 0.
    zero_point = qmin - calibration['min'] / scale
    zero_point = numerics.convert_to(zero_point, qtype)
  elif 'absmax' in calibration:
    qmax = numerics.get_symmetric_bound(qtype)
    scale = calibration['absmax'] / qmax
    # Maybe adding an epsilon (1e-7) is faster?
    scale = jnp.where(scale == 0, 1, scale)  # Scale shouldn't be 0.
    zero_point = None
  else:
    raise ValueError(f'Unsupported calibration: {calibration}')
  if qtype == 'mxfp8' or qtype == 'mxfp4':
    log2_scale = jnp.ceil(jnp.log2(scale))
    scale = (2**log2_scale).astype(scale.dtype)
  return scale, zero_point


def quantize_with_scale_zero_point(
    array: jax.Array,
    qtype: jax.typing.DTypeLike,
    scale: jax.Array,
    zero_point: jax.Array | None,
    noise_fn: numerics.NoiseFn | None = None,
) -> QArray:
  """Quantizes an array with the given scale and zero_point.

  Args:
    array: The array to quantize.
    qtype: The logical type used for quantization.
    scale: The scale to use.
    zero_point: The zero_point to use.
    noise_fn: The noise function to add to the quantized array for stochastic
      rounding.

  Returns:
    The quantized array.
  """
  if not numerics.should_quantize(array.dtype):
    raise ValueError(f'Refuse to quantize: {array.dtype}')
  if zero_point is not None and zero_point.shape != scale.shape:
    raise ValueError(
        f'Expect zero_point shape {scale.shape} but got {zero_point.shape}'
    )

  # Ensure that the scale has the same dtype as the fp array, because
  # dequantize() uses the scale dtype to reconstruct the original array.
  scale = scale.astype(array.dtype)

  qvalue = call_with_generic_broadcast(jnp.divide, array, scale)
  if zero_point is not None:
    qvalue = call_with_generic_broadcast(
        jnp.add, qvalue, zero_point.astype(qvalue.dtype)
    )
  qvalue = numerics.convert_to(qvalue, qtype, noise_fn)
  return QArray(qvalue, scale, zero_point, qtype)


def quantize(array: jax.Array, how: HowToQuantize) -> QArray:
  """Quantizes an array using a dynamic range."""
  calibration = calibrate(array, how)
  scale, zero_point = compute_scale_zero_point(calibration, how.qtype)
  return quantize_with_scale_zero_point(
      array, how.qtype, scale, zero_point, how.noise_fn
  )


def quantize_api(
    array: jax.Array,
    qtype: jax.typing.DTypeLike,
    *,
    channelwise_axes: Collection[int] = (),
    tiled_axes: Mapping[int, int | float] | None = None,
    calibration_method: str = 'absmax',
    scale_dtype: jax.typing.DTypeLike | None = None,
) -> QArray:
  """Quantize a Jax Array into QArray using a dynamic range.

  This function exists as a public API to avoid constructing a HowToQuantize.

  Args:
    array: The array to quantize.
    qtype: The logical type of the quantized value, e.g. jnp.int8, jnp.int4,
      jnp.float8_e4m3fn, "nf4", etc.
    channelwise_axes: Channelwise axes have individual scales. This has the same
      effect as setting their tile sizes to 1 in tiled_axes.
    tiled_axes: Tiled axes have blockwise scales, aka subchannel quantization.
      The value is a mapping from the tiled axis to the tile size. If the tile
      size is a float, it will be interpreted as "1 / tile_count" and the actual
      tile size will be round(axis_size * tile_size).
    calibration_method: The calibration method to use. The format is
      "<method>[,<args>]", e.g. "absmax" or "fixed,-10,10".
    scale_dtype: The dtype of the scale. If not given, the dtype will be the
      same as the array's dtype. Note that the scale's dtype decides the
      dequantized array's dtype.

  Returns:
    The quantized array.
  """
  # A stable API for qarray.quantize()
  how = HowToQuantize(
      qtype=qtype,
      channelwise_axes=channelwise_axes,
      tiled_axes=tiled_axes or {},
      calibration_method=calibration_method,
  )
  array = quantize(array, how)
  if scale_dtype is not None:
    array = dataclasses.replace(array, scale=array.scale.astype(scale_dtype))
  return array


def dequantize(array: QArray) -> jax.Array:
  """Dequantizes an array. The reverse of |quantize|.

  Args:
    array: The quantized array to dequantize.

  Returns:
    The dequantized array, whose dtype is the same as the scale's dtype.
  """
  validate_qarray(array)
  qvalue = numerics.convert_from(array.qvalue, array.qtype)
  qvalue = qvalue.astype(array.scale.dtype)
  if array.zero_point is not None:
    qvalue = call_with_generic_broadcast(
        jnp.subtract, qvalue, array.zero_point.astype(qvalue.dtype)
    )
  return call_with_generic_broadcast(jnp.multiply, qvalue, array.scale)


def clip_to_calibration(
    array: jax.Array,
    calibration: Mapping[str, jax.Array],
    tiled_axes: Mapping[int, int | float],
) -> jax.Array:
  """Clips an array to the calibration range."""
  original_shape = array.shape
  array = split_axis(array, tiled_axes)
  if 'min' in calibration and 'max' in calibration:
    min_array = split_axis(calibration['min'], {a: 1 for a in tiled_axes})
    max_array = split_axis(calibration['max'], {a: 1 for a in tiled_axes})
    array = jnp.clip(array, min_array, max_array)
  elif 'absmax' in calibration:
    absmax = split_axis(calibration['absmax'], {a: 1 for a in tiled_axes})
    array = jnp.clip(array, -absmax, absmax)
  else:
    raise ValueError(f'Unsupported calibration: {calibration}')
  return array.reshape(original_shape)


def get_accumulator_and_result_type(
    *args: MaybeQArray,
    preferred_element_type: jax.typing.DTypeLike | None,
) -> tuple[jax.typing.DTypeLike, jax.typing.DTypeLike]:
  """jnp.result_type for QArray.

  Accumulator type is the dtype used for the dot_general computation.
  Result type is the dtype of the final result.

  Args:
    *args: The arguments to dot_general.
    preferred_element_type: The preferred element type for dot_general.

  Returns:
    A tuple of the accumulator type and the result type.
  """
  qvalue_dtypes, dequant_dtypes = [], []
  for arg in args:
    if isinstance(arg, QArray):
      qvalue_dtypes.append(arg.qvalue.dtype)  # note qtype can be different.
      dequant_dtypes.append(arg.scale.dtype)
    else:
      qvalue_dtypes.append(arg.dtype)
      dequant_dtypes.append(arg.dtype)

  # Result type should only depend on dequant_dtype and preferred_element_type.
  result_type = preferred_element_type
  if result_type is None:
    # There's no dtype promotion path for fp8 or lower, and int4 or lower.
    # We manually upcast them to bf16 or int32.
    for i, t in enumerate(dequant_dtypes):
      if t.itemsize <= 1:
        dequant_dtypes[i] = jnp.int32 if 'int' in t.name else jnp.bfloat16
    result_type = jnp.result_type(*dequant_dtypes)

  # Accumulator type should be the same as result type except for int x int.
  accumulator_type = result_type
  if all('int' in t.name for t in qvalue_dtypes):
    accumulator_type = jnp.int32

  return accumulator_type, result_type



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Stochastic rounding utilities."""

from typing import Callable, Sequence
import jax
import jax.numpy as jnp


def low_bit_uniform_noise(
    key: jax.Array,
    shape: tuple[int, ...],
) -> jax.Array:
  """Random float32 numbers in nearly (-0.5, 0.5) of shape `shape`.

  Shifts the bits to the leading mantissa bit of the `f32` representation,
  and binary ORs with `1.0`, resulting in a normal number between
  `[1.0, 1.0 + (x - 1) / x]`, where `x = 2 ** rand_bits`.

  From there we adjust to `[-0.5 + 1/(2x), 0.5 - 1/(2x)]`, but don't bother
  with scaling up (converges fine for 8bits+).

  Args:
    key: Random key.
    shape: Shape of the noise.

  Returns:
    A random float32 array of shape `shape` in range (-0.5, 0.5).
  """
  nbits = 8
  rand_u8 = jax.random.bits(key, shape, dtype=jnp.uint8)

  dtype = jnp.float32
  nmant = jnp.finfo(dtype).nmant
  u_type = jnp.uint32

  rand = rand_u8.astype(u_type)
  if nmant > nbits:
    r_bitpattern = jnp.left_shift(rand, nmant - nbits)
  else:
    r_bitpattern = jnp.right_shift(rand, nbits - nmant)

  one_bitpattern = jax.lax.bitcast_convert_type(
      jnp.array(1.0, dtype=dtype), u_type
  )
  r_bitpattern = jnp.bitwise_or(one_bitpattern, r_bitpattern)

  rand_floats = jax.lax.bitcast_convert_type(r_bitpattern, dtype)

  shift = 2 ** (-1 - nbits)
  centered = rand_floats - (1.5 - shift)
  return centered


def uniform_noise(
    key: jax.Array,
    shape: tuple[int, ...],
) -> jax.Array:
  """Uniform noise."""
  return jax.random.uniform(key, shape) - 0.5


def get_noise_fn(
    method: str,
    key: jax.Array,
    channelwise_noise_axes: Sequence[int] = (0,),
) -> Callable[[tuple[int, ...]], jax.Array]:
  """Returns a noise function for stochastic rounding."""
  if method == 'uniform':
    fn = uniform_noise
  elif method == 'low_bit_uniform':
    fn = low_bit_uniform_noise
  else:
    raise ValueError(f'Unsupported stochastic rounding method: {method}')

  def noise_fn(shape: tuple[int, ...]) -> jax.Array:
    # Apply channelwise_noise_axes to get the noise shape. This significantly
    # reduces the overhead of creating a full noise array.
    noise_shape = tuple(
        dim if axis in channelwise_noise_axes else 1
        for axis, dim in enumerate(shape)
    )
    return fn(key, noise_shape)

  return noise_fn



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Quantized jax.lax.ragged_dot with quantized back propagation support."""

import dataclasses
import functools
import jax
from qwix._src import interception
from qwix._src.core import qarray
from qwix._src.core import ragged_dot


@dataclasses.dataclass(slots=True, frozen=True, kw_only=True)
class RaggedDotQtConfig:
  """Configuration for ragged_dot_qt."""

  # Forward pass settings
  lhs_qtype: jax.typing.DTypeLike | None = None
  rhs_qtype: jax.typing.DTypeLike | None = None

  # Backward pass settings
  dlhs_grad_qtype: jax.typing.DTypeLike | None = None
  drhs_grad_qtype: jax.typing.DTypeLike | None = None


@interception.disable_interceptions
def ragged_dot_qt_fwd(
    lhs: jax.Array,
    rhs: jax.Array,
    group_sizes: jax.Array,
    config: RaggedDotQtConfig,
    precision: jax.lax.PrecisionLike = None,
    preferred_element_type: jax.typing.DTypeLike | None = None,
    group_offset: jax.Array | None = None,
):
  """Forward pass for ragged_dot_qt custom VJP."""
  qlhs = qarray.quantize(
      # lhs shape [M, K]: contracting axis=1, channelwise axis=0
      lhs,
      qarray.HowToQuantize(qtype=config.lhs_qtype, channelwise_axes=[0]),
  )
  qrhs = qarray.quantize(
      # rhs shape [G, K, N]: contracting axis=1, channelwise axes=2
      rhs,
      qarray.HowToQuantize(qtype=config.rhs_qtype, channelwise_axes=[2]),
  )
  primal_out = ragged_dot.ragged_dot(
      qlhs, qrhs, group_sizes, precision, preferred_element_type, group_offset
  )
  return primal_out, (qlhs, qrhs, group_sizes)


def ragged_dot_qt_bwd(
    # Nondiff args passed from fwd pass
    config: RaggedDotQtConfig,
    precision: jax.lax.PrecisionLike,
    preferred_element_type: jax.typing.DTypeLike | None,
    group_offset: jax.Array | None,
    # Residuals from fwd pass
    residuals: tuple[qarray.MaybeQArray, qarray.MaybeQArray, jax.Array],
    g: jax.Array,
) -> tuple[jax.Array, jax.Array, None]:
  """Backward pass for ragged_dot_qt custom VJP."""
  (lhs, rhs, group_sizes) = residuals  # lhs [M, K], rhs [G, K, N], g [M, N]

  # dlhs = ragged_dot(g, rhs.swapaxes(1, 2))
  # [M, K] = [M, N] @ [G, N, K]
  # Transpose rhs as chlo.ragged_dot supports limited DimensionNumbers variants.
  g_for_dlhs = g
  rhs = rhs.swapaxes(1, 2)  # [G, N, K]
  if config.dlhs_grad_qtype:
    if isinstance(rhs, qarray.QArray):
      assert rhs.zero_point is None
      # Support channelwise only on non-group and non-contracting axes.
      g_for_dlhs = g * rhs.scale.squeeze(axis=2)  # [M, N] * [1, N]
      rhs = rhs.qvalue
    g_how = qarray.HowToQuantize(
        qtype=config.dlhs_grad_qtype,
        channelwise_axes=[0],  # [M, N]
    )
    g_for_dlhs = qarray.quantize(g_for_dlhs, g_how)
  dlhs = ragged_dot.ragged_dot(
      g_for_dlhs,
      rhs,
      group_sizes,
      precision=precision,
      preferred_element_type=preferred_element_type,
      group_offset=group_offset,
  )

  # drhs = ragged_dot_general(lhs, g)
  # [G, K, N] = [M, K] @ [M, N]
  drhs_dnums = jax.lax.RaggedDotDimensionNumbers(
      dot_dimension_numbers=(((0,), (0,)), ((), ())),
      lhs_ragged_dimensions=[0],
      rhs_group_dimensions=[],
  )
  g_for_drhs = g
  if config.drhs_grad_qtype:
    if isinstance(lhs, qarray.QArray):
      assert lhs.zero_point is None
      g_for_drhs = g * lhs.scale  # [M, N] * [M, 1]
      lhs = lhs.qvalue
    g_how = qarray.HowToQuantize(
        qtype=config.drhs_grad_qtype,
        channelwise_axes=[1],  # [M, N]
    )
    g_for_drhs = qarray.quantize(g_for_drhs, g_how)
  drhs = ragged_dot.ragged_dot_general(
      lhs,
      g_for_drhs,
      group_sizes,
      dimension_numbers=drhs_dnums,
      precision=precision,
      preferred_element_type=preferred_element_type,
      group_offset=group_offset,
  )

  return dlhs, drhs, None


@functools.partial(jax.custom_vjp, nondiff_argnums=(3, 4, 5, 6))
def ragged_dot_qt(
    lhs: jax.Array,
    rhs: jax.Array,
    group_sizes: jax.Array,
    config: RaggedDotQtConfig,
    precision: jax.lax.PrecisionLike = None,
    preferred_element_type: jax.typing.DTypeLike | None = None,
    group_offset: jax.Array | None = None,
) -> jax.Array:
  """Quantized ragged_dot with backpropagation support."""
  result, _ = ragged_dot_qt_fwd(
      lhs,
      rhs,
      group_sizes,
      config,
      precision,
      preferred_element_type,
      group_offset,
  )
  return result


ragged_dot_qt.defvjp(ragged_dot_qt_fwd, ragged_dot_qt_bwd)



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# You may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Quantized jax.lax.conv_general_dilated with quantized backpropagation support."""

from collections.abc import Sequence
import dataclasses
import functools
import operator
from typing import Any, Callable

import jax
import numpy as np
from qwix._src import interception
from qwix._src.core import conv_general
from qwix._src.core import numerics
from qwix._src.core import qarray


@dataclasses.dataclass(slots=True, frozen=True, kw_only=True)
class ConvGeneralQtConfig:
  """Configuration for conv_general_qt."""

  # Forward pass.
  lhs_qtype: jax.typing.DTypeLike | None = None
  rhs_qtype: jax.typing.DTypeLike | None = None
  lhs_calibration_method: str = 'absmax'
  rhs_calibration_method: str = 'absmax'
  lhs_collect_quant_stat: Callable[[Any], Any] | None = None
  rhs_collect_quant_stat: Callable[[Any], Any] | None = None

  # Backward pass.
  dlhs_grad_qtype: jax.typing.DTypeLike | None = None
  dlhs_grad_calibration_method: str = 'absmax'
  drhs_grad_qtype: jax.typing.DTypeLike | None = None
  drhs_grad_calibration_method: str = 'absmax'

  # Misc.
  disable_channelwise_axes: bool = False


# Swaps the first two dimension indices of a specification.
# e.g., for an (N, C, H, W) spec, returns a (C, N, H, W) spec.
_conv_spec_transpose = lambda spec: (spec[1], spec[0]) + tuple(spec[2:])

# Extracts the spatial dimension indices from a specification.
# e.g., for an (N, C, H, W) spec, returns the (H, W) spec.
_conv_sdims = lambda spec: spec[2:]


def _compute_dilated_shape(
    shape: Sequence[int], dilation: Sequence[int]
) -> tuple[int, ...]:
  """Computes the shape of an array after dilation."""
  if not dilation or all(d == 1 for d in dilation):
    return tuple(shape)
  return tuple(
      s + (s - 1) * (d - 1) if s > 0 else 0 for s, d in zip(shape, dilation)
  )


def _conv_general_vjp_lhs_padding(
    in_shape,
    window_dimensions,
    window_strides,
    out_shape,
    padding,
    lhs_dilation,
    rhs_dilation,
) -> list[tuple[int, int]]:
  """Computes padding for the VJP wrt the LHS of a convolution."""
  lhs_dilated_shape = _compute_dilated_shape(in_shape, lhs_dilation)
  rhs_dilated_shape = _compute_dilated_shape(window_dimensions, rhs_dilation)
  out_dilated_shape = _compute_dilated_shape(out_shape, window_strides)
  pad_before = np.subtract(rhs_dilated_shape, [lo for lo, _ in padding]) - 1
  pad_after = (
      np.add(lhs_dilated_shape, rhs_dilated_shape)
      - 1
      - out_dilated_shape
      - pad_before
  )
  return list(zip(pad_before, pad_after))


def _conv_general_vjp_rhs_padding(
    in_shape,
    window_dimensions,
    window_strides,
    out_shape,
    padding,
    lhs_dilation,
    rhs_dilation,
):
  """Computes padding for the VJP wrt the RHS of a convolution."""
  if not in_shape:  # 0D conv
    return []
  lhs_dilated_shape = _compute_dilated_shape(in_shape, lhs_dilation)
  rhs_dilated_shape = _compute_dilated_shape(window_dimensions, rhs_dilation)
  out_dilated_shape = _compute_dilated_shape(out_shape, window_strides)
  pads_lo, _ = zip(*padding)
  pads_from_lhs = map(operator.sub, out_dilated_shape, lhs_dilated_shape)
  pads_from_rhs = tuple(
      rd - pd - 1 for rd, pd in zip(rhs_dilated_shape, pads_lo)
  )
  pads_hi = tuple(map(operator.add, pads_from_lhs, pads_from_rhs))
  return list(zip(pads_lo, pads_hi))


def _apply_fwd_scale_to_g(
    scale: jax.Array, g: jax.Array, g_axis: int
) -> jax.Array:
  """Applies the scale from the forward pass to the backward gradient.

  Since the scale can only have channelwise axes on one dimension, we only need
  to know which axis in g this scale applies to.

  Args:
    scale: The scale from the forward pass. Only 1 axis can be channelwise.
    g: The backward gradient.
    g_axis: The axis in g that the scale applies to.

  Returns:
    The backward gradient with the scale applied.
  """
  scale = scale.reshape([-1 if a == g_axis else 1 for a in range(g.ndim)])
  return g * scale


@interception.disable_interceptions
def conv_general_qt_fwd(
    lhs: jax.Array,
    rhs: jax.Array,
    config: ConvGeneralQtConfig,
    window_strides: Sequence[int],
    padding: str | Sequence[tuple[int, int]],
    lhs_dilation: Sequence[int] | None,
    rhs_dilation: Sequence[int] | None,
    dimension_numbers: jax.lax.ConvDimensionNumbers | None,
    feature_group_count: int,
    batch_group_count: int,
) -> tuple[jax.Array, tuple[qarray.MaybeQArray, qarray.MaybeQArray]]:
  """Forward pass for conv_general_qt custom VJP."""
  dnums = jax.lax.conv_dimension_numbers(
      lhs.shape, rhs.shape, dimension_numbers
  )

  def _quantize_operand(
      operand: jax.Array, *, for_lhs: bool
  ) -> qarray.MaybeQArray:
    """Quantizes a single operand for the forward pass if configured to do so."""
    qtype = config.lhs_qtype if for_lhs else config.rhs_qtype
    if not (qtype and numerics.should_quantize(operand.dtype)):
      return operand

    if for_lhs:
      calibration_method = config.lhs_calibration_method
      collect_quant_stat = config.lhs_collect_quant_stat
    else:
      calibration_method = config.rhs_calibration_method
      collect_quant_stat = config.rhs_collect_quant_stat

    how = conv_general.get_how_to_quantize(
        dimension_numbers=dnums,
        for_lhs=for_lhs,
        qtype=qtype,
        calibration_method=calibration_method,
    )
    if config.disable_channelwise_axes:
      how = dataclasses.replace(how, channelwise_axes=[])

    calibration = qarray.calibrate(operand, how)
    if collect_quant_stat:
      calibration = collect_quant_stat(calibration)
    scale, zero_point = qarray.compute_scale_zero_point(calibration, qtype)
    return qarray.quantize_with_scale_zero_point(
        operand, qtype, scale, zero_point
    )

  lhs = _quantize_operand(lhs, for_lhs=True)
  rhs = _quantize_operand(rhs, for_lhs=False)

  primal_out = conv_general.conv_general_dilated(
      lhs,
      rhs,
      window_strides,
      padding,
      lhs_dilation,
      rhs_dilation,
      dnums,
      feature_group_count,
      batch_group_count,
  )
  residuals = (lhs, rhs)

  return primal_out, residuals


def conv_general_qt_bwd(
    config: ConvGeneralQtConfig,
    window_strides: Sequence[int],
    padding: str | Sequence[tuple[int, int]],
    lhs_dilation: Sequence[int] | None,
    rhs_dilation: Sequence[int] | None,
    dimension_numbers: jax.lax.ConvDimensionNumbers | None,
    feature_group_count: int,
    batch_group_count: int,
    res: tuple[qarray.MaybeQArray, qarray.MaybeQArray],
    g: jax.Array,
):
  """Backward pass for conv_general_qt custom VJP."""
  lhs, rhs = res

  dnums = jax.lax.conv_dimension_numbers(
      lhs.shape, rhs.shape, dimension_numbers
  )
  if lhs_dilation is None:
    lhs_dilation = (1,) * (lhs.ndim - 2)
  if rhs_dilation is None:
    rhs_dilation = (1,) * (rhs.ndim - 2)

  lhs_sdims_indices, rhs_sdims_indices, out_sdims_indices = map(
      _conv_sdims, dnums
  )
  lhs_sdims_shape = tuple(np.take(lhs.shape, lhs_sdims_indices))
  rhs_sdims_shape = tuple(np.take(rhs.shape, rhs_sdims_indices))
  g_sdims_shape = tuple(np.take(g.shape, out_sdims_indices))

  if isinstance(padding, str):
    effective_rhs_sdims = _compute_dilated_shape(rhs_sdims_shape, rhs_dilation)
    padding = jax.lax.padtype_to_pads(
        lhs_sdims_shape, effective_rhs_sdims, window_strides, padding
    )

  # dlhs
  dlhs_dnums = jax.lax.ConvDimensionNumbers(
      lhs_spec=dnums.out_spec,
      rhs_spec=_conv_spec_transpose(dnums.rhs_spec),
      out_spec=dnums.lhs_spec,
  )
  dlhs_padding = _conv_general_vjp_lhs_padding(
      lhs_sdims_shape,
      rhs_sdims_shape,
      window_strides,
      g_sdims_shape,
      padding,
      lhs_dilation,
      rhs_dilation,
  )

  dlhs_g = g
  if config.dlhs_grad_qtype:
    if isinstance(rhs, qarray.QArray):
      # Apply rhs.scale to dlhs_g.
      dlhs_g = _apply_fwd_scale_to_g(rhs.scale, dlhs_g, dnums.out_spec[1])
      rhs = rhs.qvalue
    how = conv_general.get_how_to_quantize(
        dimension_numbers=dlhs_dnums,
        for_lhs=True,
        qtype=config.dlhs_grad_qtype,
        calibration_method=config.dlhs_grad_calibration_method,
    )
    if config.disable_channelwise_axes:
      how = dataclasses.replace(how, channelwise_axes=[])
    dlhs_g = qarray.quantize(dlhs_g, how)

  rhs = jax.tree.map(lambda x: jax.lax.rev(x, rhs_sdims_indices), rhs)

  dlhs = conv_general.conv_general_dilated(
      lhs=dlhs_g,
      rhs=rhs,
      window_strides=lhs_dilation,
      padding=dlhs_padding,
      lhs_dilation=window_strides,
      rhs_dilation=rhs_dilation,
      dimension_numbers=dlhs_dnums,
      feature_group_count=feature_group_count,
      batch_group_count=batch_group_count,
  )

  # drhs
  drhs_dnums = jax.lax.ConvDimensionNumbers(
      lhs_spec=_conv_spec_transpose(dnums.lhs_spec),
      rhs_spec=_conv_spec_transpose(dnums.out_spec),
      out_spec=_conv_spec_transpose(dnums.rhs_spec),
  )
  drhs_padding = _conv_general_vjp_rhs_padding(
      lhs_sdims_shape,
      rhs_sdims_shape,
      window_strides,
      g_sdims_shape,
      padding,
      lhs_dilation,
      rhs_dilation,
  )

  drhs_g = g
  if config.drhs_grad_qtype:
    if isinstance(lhs, qarray.QArray):
      # Apply lhs.scale to drhs_g.
      drhs_g = _apply_fwd_scale_to_g(lhs.scale, drhs_g, dnums.out_spec[0])
      lhs = lhs.qvalue
    how = conv_general.get_how_to_quantize(
        dimension_numbers=drhs_dnums,
        for_lhs=False,
        qtype=config.drhs_grad_qtype,
        calibration_method=config.drhs_grad_calibration_method,
    )
    if config.disable_channelwise_axes:
      how = dataclasses.replace(how, channelwise_axes=[])
    drhs_g = qarray.quantize(drhs_g, how)

  drhs = conv_general.conv_general_dilated(
      lhs=lhs,
      rhs=drhs_g,
      window_strides=rhs_dilation,
      padding=drhs_padding,
      lhs_dilation=lhs_dilation,
      rhs_dilation=window_strides,
      dimension_numbers=drhs_dnums,
      feature_group_count=feature_group_count,
      batch_group_count=batch_group_count,
  )

  return dlhs, drhs


@functools.partial(jax.custom_vjp, nondiff_argnums=(2, 3, 4, 5, 6, 7, 8, 9))
def conv_general_qt(
    lhs: jax.Array,
    rhs: jax.Array,
    config: ConvGeneralQtConfig,
    window_strides: Sequence[int],
    padding: str | Sequence[tuple[int, int]],
    lhs_dilation: Sequence[int] | None = None,
    rhs_dilation: Sequence[int] | None = None,
    dimension_numbers: jax.lax.ConvDimensionNumbers | None = None,
    feature_group_count: int = 1,
    batch_group_count: int = 1,
) -> jax.Array:
  """Quantized conv_general using a simple, hashable config dataclass."""
  result, _ = conv_general_qt_fwd(
      lhs,
      rhs,
      config,
      window_strides,
      padding,
      lhs_dilation,
      rhs_dilation,
      dimension_numbers,
      feature_group_count,
      batch_group_count,
  )
  return result


conv_general_qt.defvjp(conv_general_qt_fwd, conv_general_qt_bwd)



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""QArray implementation with padding to tile sizes.

Extends qwix.QArray with PaddedQArray for automatic padding to tile size
multiples before quantization. Provides wrappers for dot_general and PTQ
functions.
"""

from __future__ import annotations

import dataclasses
import functools
import sys
from typing import Any, Mapping, TypeAlias

import flax.struct
import jax
import jax.numpy as jnp
from qwix._src.core import dot_general as core_dot_general
from qwix._src.core import einsum as core_einsum
from qwix._src.core import qarray
from qwix._src.providers import ptq as _ptq


PtqProvider = _ptq.PtqProvider
calibrate = qarray.calibrate
HowToQuantize = qarray.HowToQuantize

# Whether to keep shape of qvalue in padded form during "quantize"
QARRAY_KEEP_PADDED_SHAPE = False


# ---------------------------
# Padded QArray implementation
# ---------------------------


@flax.struct.dataclass
class PaddedQArray(qarray.QArray):
  """Quantized array with padding support.

  The qvalue stored can be either padded or unpadded, i.e.
  qvalue.shape == padded_shape or qvalue.shape == original_shape

  Attributes:
    padded_shape: The shape after padding to tile boundaries.
    original_shape: The original shape before padding.
  """

  padded_shape: tuple[int, ...] = flax.struct.field(
      pytree_node=False, default=()
  )
  original_shape: tuple[int, ...] = flax.struct.field(
      pytree_node=False, default=()
  )


MaybeQArray: TypeAlias = jax.Array | qarray.QArray | PaddedQArray


def pad_to_shape(array: jax.Array, target_shape: tuple[int, ...]) -> jax.Array:
  """Pads array to target shape."""
  if array.shape == target_shape:
    return array
  pad_width = [
      (0, target - current)
      for current, target in zip(array.shape, target_shape)
  ]
  return jnp.pad(array, pad_width, constant_values=0)


def get_padded_shape(
    original_shape: tuple[int, ...],
    tiled_axes: Mapping[int, int | float],
) -> tuple[int, ...]:
  """Computes the target padded shape given tiled axes.

  For each axis in tiled_axes, pads the dimension up to the next multiple of
  the tile size. If the tile size is a float, it follows the QWIX convention
  of interpreting it as 1 / tile_count, i.e., tile_size = round(dim * tile).

  Args:
    original_shape: The original shape of the array.
    tiled_axes: A mapping from axis index to the tile size or tile count.

  Returns:
    The shape after padding.
  """
  if not tiled_axes:
    return tuple(original_shape)
  out = list(original_shape)
  for axis, tile in tiled_axes.items():
    dim = original_shape[axis]
    tile_size = round(dim * tile) if isinstance(tile, float) else int(tile)
    if tile_size <= 0:
      continue
    remainder = dim % tile_size
    if remainder:
      out[axis] = dim + (tile_size - remainder)
  return tuple(out)


def quantize(array: jax.Array, how: HowToQuantize) -> PaddedQArray:
  """Quantizes an array using a dynamic range with padding support."""
  original_shape = array.shape
  array = pad_to_shape(array, get_padded_shape(array.shape, how.tiled_axes))
  padded_shape = array.shape
  array = qarray.quantize(array, how)
  if not QARRAY_KEEP_PADDED_SHAPE:
    array = dataclasses.replace(
        array,
        qvalue=array.qvalue[tuple(slice(0, dim) for dim in original_shape)],
    )
  return PaddedQArray(
      **dataclasses.asdict(array),
      padded_shape=padded_shape,
      original_shape=original_shape,
  )


def dequantize(array: PaddedQArray) -> jax.Array:
  """Dequantizes an array. The reverse of |quantize|."""
  qarray.validate_qarray(array)
  padded_qvalue = pad_to_shape(array.qvalue, array.padded_shape)
  out = qarray.dequantize(dataclasses.replace(array, qvalue=padded_qvalue))
  if out.shape != array.original_shape:
    out = out[tuple(slice(0, d) for d in array.original_shape)]
  return out


# ---------------------------
# dot_general and einsum wrappers
# ---------------------------


def _pad_operand_if_qarray(x):
  if isinstance(x, PaddedQArray):
    padded_q = pad_to_shape(x.qvalue, x.padded_shape)
    return dataclasses.replace(x, qvalue=padded_q)
  return x


def dot_general(
    lhs: MaybeQArray,
    rhs: MaybeQArray,
    dimension_numbers: jax.lax.DotDimensionNumbers,
    precision: jax.lax.PrecisionLike = None,
    preferred_element_type: jax.typing.DTypeLike | None = None,
    **kwargs,
) -> jax.Array:
  """Pad operands and delegate to core dot_general.

  PaddedQArray operands are padded to their stored padded_shape. Regular arrays
  are padded along contraction dimensions to match PaddedQArray operands.

  Args:
    lhs: Left-hand side operand.
    rhs: Right-hand side operand.
    dimension_numbers: Dimension numbers for dot_general.
    precision: Optional precision.
    preferred_element_type: Optional element type.
    **kwargs: Additional arguments.

  Returns:
    Result of dot_general operation.
  """

  # Pad PaddedQArray operands to their padded_shape
  lhs = _pad_operand_if_qarray(lhs)
  rhs = _pad_operand_if_qarray(rhs)

  # If only one operand is PaddedQArray, pad the other to match along
  # contraction dims
  (lhs_contract, rhs_contract), _ = dimension_numbers

  if not isinstance(rhs, PaddedQArray):
    target_shape = list(rhs.shape)
    for rhs_axis, lhs_axis in zip(rhs_contract, lhs_contract):
      target_shape[rhs_axis] = lhs.shape[lhs_axis]
    rhs = pad_to_shape(rhs, tuple(target_shape))

  if not isinstance(lhs, PaddedQArray):
    target_shape = list(lhs.shape)
    for lhs_axis, rhs_axis in zip(lhs_contract, rhs_contract):
      target_shape[lhs_axis] = rhs.shape[rhs_axis]
    lhs = pad_to_shape(lhs, tuple(target_shape))

  return core_dot_general.dot_general(
      lhs,
      rhs,
      dimension_numbers,
      preferred_element_type=preferred_element_type,
      precision=precision,
      **kwargs,
  )


def einsum(
    einsum_str: str,
    lhs: MaybeQArray,
    rhs: MaybeQArray,
    *,
    preferred_element_type: jax.typing.DTypeLike | None = None,
    **kwargs,
) -> jax.Array:
  """Pad operands and delegate to core einsum.

  PaddedQArray operands are padded to their stored padded_shape. Regular arrays
  are padded along contraction dimensions to match PaddedQArray operands.

  Args:
    einsum_str: The einsum equation string.
    lhs: Left-hand side operand.
    rhs: Right-hand side operand.
    preferred_element_type: Optional element type.
    **kwargs: Additional arguments.

  Returns:
    Result of the einsum operation.
  """

  # Pad PaddedQArray operands to their padded_shape
  lhs = _pad_operand_if_qarray(lhs)
  rhs = _pad_operand_if_qarray(rhs)

  # If only one operand is PaddedQArray, pad the other to match along
  # contraction dims
  info = core_einsum.get_einsum_info(einsum_str, (lhs.ndim, rhs.ndim))
  if not isinstance(rhs, PaddedQArray):
    target_shape = list(rhs.shape)
    for axis, name in enumerate(info.rhs):
      if name in info.contractions:
        lhs_axis = info.lhs.index(name)
        target_shape[axis] = lhs.shape[lhs_axis]
    rhs = pad_to_shape(rhs, tuple(target_shape))

  if not isinstance(lhs, PaddedQArray):
    target_shape = list(lhs.shape)
    for axis, name in enumerate(info.lhs):
      if name in info.contractions:
        rhs_axis = info.rhs.index(name)
        target_shape[axis] = rhs.shape[rhs_axis]
    lhs = pad_to_shape(lhs, tuple(target_shape))

  return core_einsum.einsum(
      einsum_str,
      lhs,
      rhs,
      preferred_element_type=preferred_element_type,
      **kwargs,
  )


def quantize_act(
    array: jax.Array,
    how: HowToQuantize,
    rule,
    act_name: str | None,
):
  """Wrapper to reuse PTQ.quantize_act with this module as qarray backend."""
  return _ptq.quantize_act(
      array, how, rule, act_name, _qarray_module=sys.modules[__name__]
  )


def create_quantized_param(
    name: str,
    value: jax.Array,
    how: HowToQuantize,
) -> _ptq.WithAux[qarray.QArray]:
  """Wrapper that delegates to PTQ.create_quantized_param using this backend."""
  return _ptq.create_quantized_param(
      name, value, how, _qarray_module=sys.modules[__name__]
  )


def quantize_params(
    params: Any,
    abstract_quantized_params: Any,
    quant_stats: Any = flax.core.FrozenDict(),
) -> Any:
  """Wrapper that delegates to PTQ.quantize_params using this backend."""
  return _ptq.quantize_params(
      params,
      abstract_quantized_params,
      quant_stats,
      _qarray_module=sys.modules[__name__],
  )


# ---------------------------
# Provider
# ---------------------------

PaddedPtqProvider = functools.partial(
    PtqProvider,
    _qarray_module=sys.modules[__name__],
    _dot_general_fn=dot_general,
    _einsum_fn=einsum,
)


__all__ = [
    'PaddedPtqProvider',
    'PaddedQArray',
]



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Quantized jax.numpy.dot with subchannel support."""

from typing import Callable

import jax
from qwix._src.core import dot_general
from qwix._src.core import qarray


def dot(
    a: qarray.MaybeQArray,
    b: qarray.MaybeQArray,
    precision: jax.lax.PrecisionLike = None,
    preferred_element_type: jax.typing.DTypeLike | None = None,
    out_sharding=None,
    *,
    _qwix_dot_general: Callable[..., jax.Array] = dot_general.dot_general,
):
  """jnp.dot with QArray support."""
  if a.ndim == 0 or b.ndim == 0:
    contract_dims = ((), ())
  elif b.ndim == 1:
    contract_dims = ((a.ndim - 1,), (0,))
  else:
    contract_dims = ((a.ndim - 1,), (b.ndim - 2,))
  return _qwix_dot_general(
      a,
      b,
      dimension_numbers=(contract_dims, ((), ())),
      precision=precision,
      preferred_element_type=preferred_element_type,
      out_sharding=out_sharding,
  )



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Quantized jax.lax.dot_general with quantized backpropagation support."""

import dataclasses
import functools
from typing import Any, Callable

import jax
from jax import numpy as jnp
import numpy as np
from qwix._src import interception
from qwix._src.core import dot_general
from qwix._src.core import numerics
from qwix._src.core import qarray


@dataclasses.dataclass(slots=True, frozen=True, kw_only=True)
class DotGeneralQtConfig:
  """Configuration for dot_general_qt."""

  # Forward pass.
  lhs_qtype: jax.typing.DTypeLike | None = None
  rhs_qtype: jax.typing.DTypeLike | None = None
  tile_size: int | float | None = None
  lhs_calibration_method: str = 'absmax'
  rhs_calibration_method: str = 'absmax'
  lhs_collect_quant_stat: Callable[[Any], Any] | None = None
  rhs_collect_quant_stat: Callable[[Any], Any] | None = None

  # Backward pass (dlhs).
  dlhs_grad_qtype: jax.typing.DTypeLike | None = None  # incoming gradient
  dlhs_grad_calibration_method: str = 'absmax'
  dlhs_tile_size: int | float | None = None
  dlhs_stochastic_rounding_noise_fn: numerics.NoiseFn | None = None

  # Backward pass (drhs).
  drhs_grad_qtype: jax.typing.DTypeLike | None = None  # incoming gradient
  drhs_grad_calibration_method: str = 'absmax'
  drhs_tile_size: int | float | None = None
  drhs_stochastic_rounding_noise_fn: numerics.NoiseFn | None = None

  # Misc.
  disable_channelwise_axes: bool = False


def _ranges_like(*xs):
  start = 0
  for x in xs:
    yield tuple(range(start, start + len(x)))
    start += len(x)


def _update_dimension_numbers_for_backward(
    fwd_dimension_numbers: jax.lax.DotDimensionNumbers,
    fwd_ndims: tuple[int, int],
    *,
    for_dlhs: bool,
) -> tuple[jax.lax.DotDimensionNumbers, tuple[int, ...]]:
  """Generates a new dimension number for backward pass.

  The dimension number is used to calculate dlhs = dot(g, rhs), if for_dlhs is
  True, or drhs = dot(g, lhs), otherwise.

  Args:
    fwd_dimension_numbers: Dimension numbers from the forward pass.
    fwd_ndims: Number of dimensions of the operands in the forward pass.
    for_dlhs: Whether the backward pass is for dlhs or drhs.

  Returns:
    A tuple of (dimension numbers for backward pass, output transpose to be
    applied to the result of the backward dot to match the original dimensions).
  """
  # We use the following conventions:
  #   g: incoming gradient.
  #   x: output gradient, dlhs if for_dlhs, drhs otherwise.
  #   y: residual, rhs if for_dlhs, lhs otherwise.
  if for_dlhs:
    (x_ca, y_ca), (x_ba, y_ba) = fwd_dimension_numbers
    x_ndim, y_ndim = fwd_ndims
  else:
    (y_ca, x_ca), (y_ba, x_ba) = fwd_dimension_numbers
    y_ndim, x_ndim = fwd_ndims

  x_ra = tuple(sorted(set(range(x_ndim)) - set(x_ca) - set(x_ba)))
  y_ra = tuple(sorted(set(range(y_ndim)) - set(y_ca) - set(y_ba)))

  if for_dlhs:
    g_ba, _, g_ca = _ranges_like(x_ba, x_ra, y_ra)
  else:
    g_ba, g_ca, _ = _ranges_like(x_ba, y_ra, x_ra)
  dnums = ((g_ca, y_ra), (g_ba, y_ba))

  x_ca_sorted_by_y = tuple(np.take(x_ca, np.argsort(y_ca)))
  out_transpose_axes = tuple(np.argsort(tuple(x_ba) + x_ra + x_ca_sorted_by_y))
  return dnums, out_transpose_axes


def _apply_rhs_scale_to_lhs(lhs, rhs_scale, dnums):
  """Applies the rhs_scale to lhs."""
  (lhs_ca, rhs_ca), (lhs_ba, rhs_ba) = dnums
  lhs_axes_to_rhs = dict(zip(lhs_ca, rhs_ca)) | dict(zip(lhs_ba, rhs_ba))
  lhs_scale = qarray.transpose_array(
      rhs_scale, [lhs_axes_to_rhs.get(a) for a in range(lhs.ndim)]
  )
  return qarray.call_with_generic_broadcast(jnp.multiply, lhs, lhs_scale)


# See test_scan_custom_vjp in interception_test.py for why we need to manually
# disable interceptions for dot_general_qt_fwd.
@interception.disable_interceptions
def dot_general_qt_fwd(
    lhs: jax.Array,
    rhs: jax.Array,
    lhs_calibration: dict[str, jax.Array] | None,
    rhs_calibration: dict[str, jax.Array] | None,
    dimension_numbers: jax.lax.DotDimensionNumbers,
    config: DotGeneralQtConfig,
):
  """Forward pass for dot_general_qt custom VJP."""
  if lhs_calibration is not None:
    scale, zero_point = qarray.compute_scale_zero_point(
        lhs_calibration, config.lhs_qtype
    )
    lhs = qarray.quantize_with_scale_zero_point(
        lhs, config.lhs_qtype, scale, zero_point
    )
  if rhs_calibration is not None:
    scale, zero_point = qarray.compute_scale_zero_point(
        rhs_calibration, config.rhs_qtype
    )
    rhs = qarray.quantize_with_scale_zero_point(
        rhs, config.rhs_qtype, scale, zero_point
    )
  return dot_general.dot_general(lhs, rhs, dimension_numbers), (lhs, rhs)


def dot_general_qt_bwd(
    fwd_dimension_numbers: jax.lax.DotDimensionNumbers,
    config: DotGeneralQtConfig,
    residuals: tuple[qarray.MaybeQArray, qarray.MaybeQArray],
    g: jax.Array,
):
  """Backward pass for dot_general_qt custom VJP."""
  lhs, rhs = residuals

  def _compute_gradient_for_operand(
      g: jax.Array, y: qarray.MaybeQArray, *, for_dlhs: bool
  ):
    """Compute dot_general for gradient and other_fwd_operand."""
    bwd_dnums, transpose_axes = _update_dimension_numbers_for_backward(
        fwd_dimension_numbers, (lhs.ndim, rhs.ndim), for_dlhs=for_dlhs
    )
    if for_dlhs:
      g_qtype = config.dlhs_grad_qtype
      g_tile_size = config.dlhs_tile_size
      g_calibration_method = config.dlhs_grad_calibration_method
      g_noise_fn = config.dlhs_stochastic_rounding_noise_fn
    else:
      g_qtype = config.drhs_grad_qtype
      g_tile_size = config.drhs_tile_size
      g_calibration_method = config.drhs_grad_calibration_method
      g_noise_fn = config.drhs_stochastic_rounding_noise_fn

    if g_qtype and numerics.should_quantize(g.dtype):
      if isinstance(y, qarray.QArray) and not qarray.get_tiled_axes(y):
        # Apply the scale of y to g, this trick avoids requantizing y because
        # the y from fwd pass has different channelwise_axes.
        assert y.zero_point is None and y.qtype == y.qvalue.dtype
        g = _apply_rhs_scale_to_lhs(g, y.scale, bwd_dnums)
        y = y.qvalue

      g_how = dot_general.get_how_to_quantize(
          dimension_numbers=bwd_dnums,
          ndims=(g.ndim, y.ndim),
          for_lhs=True,
          qtype=g_qtype,
          tile_size=g_tile_size,
          calibration_method=g_calibration_method,
          noise_fn=g_noise_fn,
      )
      if config.disable_channelwise_axes:
        g_how = dataclasses.replace(g_how, channelwise_axes=[])

      g = qarray.quantize(g, g_how)

    grad_res = dot_general.dot_general(g, y, bwd_dnums)
    return jax.lax.transpose(grad_res, transpose_axes)

  dlhs = _compute_gradient_for_operand(g, rhs, for_dlhs=True)
  drhs = _compute_gradient_for_operand(g, lhs, for_dlhs=False)

  return dlhs, drhs, None, None


@functools.partial(jax.custom_vjp, nondiff_argnums=(4, 5))
def dot_general_qt_fwd_bwd(
    lhs: jax.Array,
    rhs: jax.Array,
    lhs_calibration: dict[str, jax.Array] | None,
    rhs_calibration: dict[str, jax.Array] | None,
    dimension_numbers: jax.lax.DotDimensionNumbers,
    config: DotGeneralQtConfig,
) -> jax.Array:
  """dot_general custom VJP."""
  result, _ = dot_general_qt_fwd(
      lhs, rhs, lhs_calibration, rhs_calibration, dimension_numbers, config
  )
  return result


dot_general_qt_fwd_bwd.defvjp(dot_general_qt_fwd, dot_general_qt_bwd)


def dot_general_qt(
    lhs: jax.Array,
    rhs: jax.Array,
    dimension_numbers: jax.lax.DotDimensionNumbers,
    config: DotGeneralQtConfig,
) -> jax.Array:
  """Quantized dot_general with backpropagation support."""
  lhs_calibration = None
  rhs_calibration = None

  if config.lhs_qtype and numerics.should_quantize(lhs.dtype):
    lhs_how = dot_general.get_how_to_quantize(
        dimension_numbers=dimension_numbers,
        ndims=(lhs.ndim, rhs.ndim),
        for_lhs=True,
        qtype=config.lhs_qtype,
        tile_size=config.tile_size,
        calibration_method=config.lhs_calibration_method,
    )
    if config.disable_channelwise_axes:
      lhs_how = dataclasses.replace(lhs_how, channelwise_axes=[])
    lhs_calibration = qarray.calibrate(lhs, lhs_how)
    if config.lhs_collect_quant_stat:
      lhs_calibration = config.lhs_collect_quant_stat(lhs_calibration)

  if config.rhs_qtype and numerics.should_quantize(rhs.dtype):
    rhs_how = dot_general.get_how_to_quantize(
        dimension_numbers=dimension_numbers,
        ndims=(lhs.ndim, rhs.ndim),
        for_lhs=False,
        qtype=config.rhs_qtype,
        tile_size=config.tile_size,
        calibration_method=config.rhs_calibration_method,
    )
    if config.disable_channelwise_axes:
      rhs_how = dataclasses.replace(rhs_how, channelwise_axes=[])
    rhs_calibration = qarray.calibrate(rhs, rhs_how)
    if config.rhs_collect_quant_stat:
      rhs_calibration = config.rhs_collect_quant_stat(rhs_calibration)

  return dot_general_qt_fwd_bwd(
      lhs, rhs, lhs_calibration, rhs_calibration, dimension_numbers, config
  )



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Quantized training (QT) support."""

import dataclasses
import functools
from typing import Any, Callable, Mapping, Sequence

import jax
from jax import numpy as jnp
from qwix._src import averaging
from qwix._src import flax_util
from qwix._src import qconfig
from qwix._src.core import conv_general_qt
from qwix._src.core import dot_general_qt
from qwix._src.core import ragged_dot_qt
from qwix._src.core import stochastic_rounding


@dataclasses.dataclass(frozen=True, kw_only=True)
class QtRule(qconfig.QuantizationRule):
  """QuantizationRule with all settings specific to Quantized Training (QT)."""

  # In backward pass, quantize the gradients to the given type. This doesn't
  # affect the residuals as the residuals will reuse the quantization in the
  # forward pass.
  bwd_qtype: jax.typing.DTypeLike | None = None

  # In backward pass, calibrate the gradients using the given method.
  bwd_calibration_method: str = 'absmax'

  # In backward pass, enable subchannel for contraction axes when calculating
  # the gradient of weights. Note that the tiling is actually applied to the
  # the incoming gradient and the residual activation rather than any "weight".
  bwd_weight_grad_tile_size: int | float | None = None

  # If True, disable channelwise axes for both forward and backward passes.
  disable_channelwise_axes: bool = False

  # Use stochastic rounding for the gradients.
  # Currently supports "uniform" and "low_bit_uniform".
  bwd_stochastic_rounding: str | None = None

  # Use channelwise noise for stochastic rounding. By default, it will generate
  # noise for the 0th dimension and broadcast it over remaining dimensions.
  channelwise_noise_axes: Sequence[int] = (0,)

  # Override any fields in DotGeneralQtConfig or ConvGeneralQtConfig. This is
  # highly experimental and subjects to changes with no backward compatibility
  # guarantees.
  additional_qt_config: Mapping[str, Any] | None = None


class QtProvider(qconfig.QuantizationProvider):
  """Quantization provider for Quantized Training (QT)."""

  def _init_rule(self, rule: qconfig.QuantizationRule) -> QtRule:
    rule = super()._init_rule(rule)
    if not isinstance(rule, QtRule):
      rule = QtRule(**dataclasses.asdict(rule))
    return rule

  def dot_general(
      self,
      lhs: jax.Array,
      rhs: jax.Array,
      dimension_numbers: jax.lax.DotDimensionNumbers,
      precision: jax.lax.PrecisionLike = None,
      preferred_element_type: jax.typing.DTypeLike | None = None,
      *,
      out_sharding=None,
  ) -> jax.Array:
    """QT dot_general."""
    rule, op_id = self._get_current_rule_and_op_id('dot_general')
    if rule is None or rule.weight_qtype is None:
      return jax.lax.dot_general(
          lhs,
          rhs,
          dimension_numbers,
          precision=precision,
          preferred_element_type=preferred_element_type,
          out_sharding=out_sharding,
      )
    config = self._create_dot_general_qt_config(rule, op_id, lhs, rhs)
    return dot_general_qt.dot_general_qt(lhs, rhs, dimension_numbers, config)

  def einsum(
      self,
      einsum_str: str,
      *operands: jax.Array,
      precision: jax.lax.PrecisionLike = None,
      preferred_element_type: jax.typing.DTypeLike | None = None,
      _dot_general: Callable[..., jax.Array] = jax.lax.dot_general,  # pylint: disable=invalid-name
      out_sharding=None,
  ) -> jax.Array:
    """QT einsum."""
    rule, op_id = self._get_current_rule_and_op_id('einsum')
    if rule is None or rule.weight_qtype is None:
      return jnp.einsum(
          einsum_str,
          *operands,
          precision=precision,
          preferred_element_type=preferred_element_type,
          _dot_general=_dot_general,
          out_sharding=out_sharding,
      )
    if not isinstance(einsum_str, str) or len(operands) != 2:
      raise ValueError(f'Unsupported einsum format: {einsum_str=} {operands=}')

    def custom_dot_general(
        lhs,
        rhs,
        dimension_numbers,
        precision,
        preferred_element_type,
        **kwargs,
    ):
      # TODO(dangyi): support preferred_element_type.
      del precision, preferred_element_type, kwargs
      return dot_general_qt.dot_general_qt(
          lhs,
          rhs,
          dimension_numbers,
          # lhs and rhs might be flipped by einsum so we cannot use the operands
          # from the einsum call.
          self._create_dot_general_qt_config(rule, op_id, lhs, rhs),
      )

    with jax.disable_jit():
      return jnp.einsum(
          einsum_str,
          *operands,
          precision=precision,
          preferred_element_type=preferred_element_type,
          _dot_general=custom_dot_general,
          out_sharding=out_sharding,
      )

  def conv_general_dilated(
      self,
      lhs: jax.Array,
      rhs: jax.Array,
      window_strides: Sequence[int],
      padding: str | Sequence[tuple[int, int]],
      lhs_dilation: Sequence[int] | None = None,
      rhs_dilation: Sequence[int] | None = None,
      dimension_numbers: jax.lax.ConvGeneralDilatedDimensionNumbers = None,
      feature_group_count: int = 1,
      batch_group_count: int = 1,
      precision: jax.lax.PrecisionLike = None,
      preferred_element_type: jax.typing.DTypeLike | None = None,
  ) -> jax.Array:
    """QT conv_general_dilated."""
    rule, op_id = self._get_current_rule_and_op_id('conv_general_dilated')
    if rule is None or rule.weight_qtype is None:
      return jax.lax.conv_general_dilated(
          lhs,
          rhs,
          window_strides,
          padding,
          lhs_dilation=lhs_dilation,
          rhs_dilation=rhs_dilation,
          dimension_numbers=dimension_numbers,
          feature_group_count=feature_group_count,
          batch_group_count=batch_group_count,
          precision=precision,
          preferred_element_type=preferred_element_type,
      )
    if rule.tile_size:
      raise ValueError('subchannel is not supported for conv_general_dilated.')
    config = self._create_conv_general_qt_config(rule, op_id, lhs, rhs)
    return conv_general_qt.conv_general_qt(
        lhs,
        rhs,
        config,
        window_strides,
        padding,
        lhs_dilation,
        rhs_dilation,
        dimension_numbers,
        feature_group_count,
        batch_group_count,
    )

  def ragged_dot(
      self,
      lhs: jax.Array,
      rhs: jax.Array,
      group_sizes: jax.Array,
      precision: jax.lax.PrecisionLike = None,
      preferred_element_type: jax.typing.DTypeLike | None = None,
      group_offset: jax.Array | None = None,
  ) -> jax.Array:
    """QT ragged_dot."""
    rule, _ = self._get_current_rule_and_op_id('ragged_dot')
    if rule is None or rule.weight_qtype is None:
      return jax.lax.ragged_dot(
          lhs,
          rhs,
          group_sizes,
          precision=precision,
          preferred_element_type=preferred_element_type,
          group_offset=group_offset,
      )
    config = self._create_ragged_dot_qt_config(rule)
    return ragged_dot_qt.ragged_dot_qt(
        lhs,
        rhs,
        group_sizes,
        config,
        precision,
        preferred_element_type,
        group_offset,
    )

  def get_intercept_map(self):
    """Used for interception."""
    return super().get_intercept_map() | {
        'jax.lax.conv_general_dilated': self.conv_general_dilated,
        'jax.lax.dot_general': self.dot_general,
        'jax.numpy.einsum': self.einsum,
        'jax.lax.ragged_dot': self.ragged_dot,
    }

  def _collect_quant_stat(
      self,
      name: str,
      batch_axes: tuple[int, ...],
      calibration: averaging.Calibration,
  ) -> averaging.Calibration:
    """Collects the quantization statistics."""
    # Calculate the mean over the batch axes.
    calibration = jax.tree.map(
        lambda x: x.mean(axis=batch_axes, keepdims=True), calibration
    )

    aggregator = averaging.SimpleMovingAverage()
    quant_stat = flax_util.get_or_create_variable(
        'quant_stats', name, lambda: aggregator.init(calibration)
    )

    if flax_util.should_update_quant_stats():
      quant_stat.value = aggregator.update(quant_stat.value, calibration)

    return aggregator.get_calibration(quant_stat.value, calibration)

  def _create_conv_general_qt_config(
      self,
      rule: qconfig.QuantizationRule,
      op_id: str,
      lhs: jax.Array,
      rhs: jax.Array,
  ) -> conv_general_qt.ConvGeneralQtConfig:
    """Creates a ConvGeneralQtConfig for conv_general_dilated."""
    assert isinstance(rule, QtRule), '_init_rule should have been called.'

    # Assume LHS is an activation and RHS is a weight.
    del lhs
    lhs_collect_quant_stat = None
    if rule.act_qtype is not None and rule.act_static_scale:
      lhs_collect_quant_stat = functools.partial(
          self._collect_quant_stat, f'{op_id}_lhs', rule.act_batch_axes
      )
    assert flax_util.find_param(rhs) is not None

    return conv_general_qt.ConvGeneralQtConfig(
        # fwd configs.
        lhs_qtype=rule.act_qtype,
        rhs_qtype=rule.weight_qtype,
        lhs_calibration_method=rule.act_calibration_method,
        rhs_calibration_method=rule.weight_calibration_method,
        lhs_collect_quant_stat=lhs_collect_quant_stat,
        rhs_collect_quant_stat=None,
        # bwd configs.
        dlhs_grad_qtype=rule.bwd_qtype,
        dlhs_grad_calibration_method=rule.bwd_calibration_method,
        drhs_grad_qtype=rule.bwd_qtype,
        drhs_grad_calibration_method=rule.bwd_calibration_method,
        # misc.
        disable_channelwise_axes=rule.disable_channelwise_axes,
    )

  def _create_dot_general_qt_config(
      self,
      rule: qconfig.QuantizationRule,
      op_id: str,
      lhs: jax.Array,
      rhs: jax.Array,
  ) -> dot_general_qt.DotGeneralQtConfig:
    """Creates a DotGeneralQtConfig for dot_general and einsum."""
    assert isinstance(rule, QtRule), '_init_rule should have been called.'

    # LHS configs based on whether it's a weight or an activation.
    lhs_qtype = None
    lhs_calibration_method = None
    lhs_is_weight = flax_util.find_param(lhs) is not None
    lhs_collect_quant_stat = None

    if lhs_is_weight:
      if rule.weight_qtype is not None:
        lhs_qtype = rule.weight_qtype
        lhs_calibration_method = rule.weight_calibration_method
    elif rule.act_qtype is not None:
      lhs_qtype = rule.act_qtype
      lhs_calibration_method = rule.act_calibration_method
      if rule.act_static_scale:
        lhs_collect_quant_stat = functools.partial(
            self._collect_quant_stat, f'{op_id}_lhs', rule.act_batch_axes
        )

    # RHS configs based on whether it's a weight or an activation.
    rhs_qtype = None
    rhs_calibration_method = None
    rhs_is_weight = flax_util.find_param(rhs) is not None
    rhs_collect_quant_stat = None

    if rhs_is_weight:
      assert not lhs_is_weight, 'lhs and rhs cannot be both weights.'
      if rule.weight_qtype is not None:
        rhs_qtype = rule.weight_qtype
        rhs_calibration_method = rule.weight_calibration_method
    elif rule.act_qtype is not None:
      rhs_qtype = rule.act_qtype
      rhs_calibration_method = rule.act_calibration_method
      if rule.act_static_scale:
        rhs_collect_quant_stat = functools.partial(
            self._collect_quant_stat, f'{op_id}_rhs', rule.act_batch_axes
        )

    # bwd config, which is only enabled when bwd_qtype is set.
    dlhs_tile_size = None
    drhs_tile_size = None
    bwd_stochastic_rounding_noise_fn = None

    if rule.bwd_qtype is not None:
      if lhs_is_weight:
        dlhs_tile_size = rule.bwd_weight_grad_tile_size
      if rhs_is_weight:
        drhs_tile_size = rule.bwd_weight_grad_tile_size
      if rule.bwd_stochastic_rounding is not None:
        bwd_stochastic_rounding_noise_fn = stochastic_rounding.get_noise_fn(
            method=rule.bwd_stochastic_rounding,
            key=flax_util.make_rng('stochastic_rounding'),
            channelwise_noise_axes=rule.channelwise_noise_axes,
        )

    qt_config = dot_general_qt.DotGeneralQtConfig(
        # fwd configs.
        lhs_qtype=lhs_qtype,
        rhs_qtype=rhs_qtype,
        tile_size=rule.tile_size,
        lhs_calibration_method=lhs_calibration_method,
        rhs_calibration_method=rhs_calibration_method,
        lhs_collect_quant_stat=lhs_collect_quant_stat,
        rhs_collect_quant_stat=rhs_collect_quant_stat,
        # dlhs configs.
        dlhs_grad_qtype=rule.bwd_qtype,
        dlhs_grad_calibration_method=rule.bwd_calibration_method,
        dlhs_tile_size=dlhs_tile_size,
        dlhs_stochastic_rounding_noise_fn=bwd_stochastic_rounding_noise_fn,
        # drhs configs.
        drhs_grad_qtype=rule.bwd_qtype,
        drhs_tile_size=drhs_tile_size,
        drhs_grad_calibration_method=rule.bwd_calibration_method,
        drhs_stochastic_rounding_noise_fn=bwd_stochastic_rounding_noise_fn,
        # misc.
        disable_channelwise_axes=rule.disable_channelwise_axes,
    )

    if rule.additional_qt_config:
      qt_config = dataclasses.replace(qt_config, **rule.additional_qt_config)
    return qt_config

  def _create_ragged_dot_qt_config(
      self,
      rule: qconfig.QuantizationRule,
  ) -> ragged_dot_qt.RaggedDotQtConfig:
    """Creates a RaggedDotQtConfig for ragged_dot."""
    assert isinstance(rule, QtRule), '_init_rule should have been called.'
    # Assume LHS is an activation and RHS is a weight.
    return ragged_dot_qt.RaggedDotQtConfig(
        # fwd configs.
        lhs_qtype=rule.act_qtype,
        rhs_qtype=rule.weight_qtype,
        # bwd configs.
        dlhs_grad_qtype=rule.bwd_qtype,
        drhs_grad_qtype=rule.bwd_qtype,
    )



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Quantized jax.lax.ragged_dot and jax.lax.ragged_dot_general."""

from collections.abc import Collection, Sequence
import jax
from jax import numpy as jnp
from qwix._src.core import numerics
from qwix._src.core import qarray


# RaggedDotDimensionNumbers that specify the simple case (i.e., qwix.ragged_dot)
_BASIC_RAGGED_DOT_DIMENSION_NUMBERS = jax.lax.RaggedDotDimensionNumbers(
    dot_dimension_numbers=(((1,), (1,)), ((), ())),
    lhs_ragged_dimensions=[0],
    rhs_group_dimensions=[0],
)


def _apply_group_channelwise_scale(
    rhs_scale: jax.Array,
    lhs_shape: tuple[int, ...],
    group_sizes: jax.Array,
    dimension_numbers: jax.lax.RaggedDotDimensionNumbers,
    precision: jax.lax.PrecisionLike,
    group_offset: jax.Array | None,
) -> jax.Array:
  """Expands the group dimension of rhs_scale using a gather-like op."""
  (lhs_ca, _), _ = dimension_numbers.dot_dimension_numbers

  # Create a `jnp.ones` tensor with the same rank and layout as lhs_val,
  # but with contracting dimensions set to size 1.
  ones_shape = list(lhs_shape)
  for contracting_axis in lhs_ca:
    ones_shape[contracting_axis] = 1
  lhs_ones = jnp.ones(tuple(ones_shape), rhs_scale.dtype)

  return jax.lax.ragged_dot_general(
      lhs_ones,
      rhs_scale,
      group_sizes,
      dimension_numbers,
      precision=precision,
      group_offset=group_offset,
  )


def _apply_tiling(
    contracting_axes: Sequence[int],
    batch_axes: Sequence[int],
    tiled_axes: Collection[int],
) -> tuple[tuple[int, ...], tuple[int, ...], tuple[int, ...]]:
  """Apply tiling to dimension numbers.

  Each tiled contracting axis is split into two axes, the first being the new
  batch axis, and the second being the new contracting axis.

  Args:
    contracting_axes: The original contracting axes.
    batch_axes: The original batch axes.
    tiled_axes: The tiled axes. Must be a subset of contracting_axes.

  Returns:
    A tuple of (new_ca, new_ba, sum_axes).
  """
  new_ca = [a + sum(t <= a for t in tiled_axes) for a in contracting_axes]
  new_ba = [a + sum(t < a for t in tiled_axes) for a in batch_axes]
  # We choose to insert the tile_count axes to the end of the batch axes.
  # Alternatively, we could insert them to the beginning or to the middle,
  # as long as lhs and rhs use the same order.
  new_ba += [
      a + sum(t < a for t in tiled_axes)
      for a in contracting_axes
      if a in tiled_axes
  ]
  sum_axes = range(len(batch_axes), len(new_ba))
  return tuple(new_ca), tuple(new_ba), tuple(sum_axes)


def _ragged_get_scale_transpose(
    dimension_numbers: jax.lax.RaggedDotDimensionNumbers,
    ndims: tuple[int, int],
) -> tuple[list[int | None], list[int | None]]:
  """Calculates the transpose permutation for lhs_scale and rhs_scale."""
  (lhs_ca, rhs_ca), (lhs_ba, rhs_ba) = dimension_numbers.dot_dimension_numbers
  lhs_ragged_dims = dimension_numbers.lhs_ragged_dimensions
  rhs_group_dims = dimension_numbers.rhs_group_dimensions

  lhs_remaining_dims = sorted(
      set(range(ndims[0])) - set(lhs_ca) - set(lhs_ba) - set(lhs_ragged_dims)
  )
  rhs_remaining_dims = sorted(
      set(range(ndims[1])) - set(rhs_ca) - set(rhs_ba) - set(rhs_group_dims)
  )

  lhs_scale_transpose = (
      list(lhs_ba)
      + list(lhs_ragged_dims)
      + list(lhs_remaining_dims)
      + [None] * len(rhs_remaining_dims)
  )
  rhs_scale_transpose = (
      list(rhs_ba)
      + [None] * (len(lhs_ragged_dims) + len(lhs_remaining_dims))
      + list(rhs_remaining_dims)
  )

  return lhs_scale_transpose, rhs_scale_transpose


def _fast_ragged_dot_general(
    lhs: qarray.MaybeQArray,
    rhs: qarray.MaybeQArray,
    group_sizes: jax.Array,
    dimension_numbers: jax.lax.RaggedDotDimensionNumbers,
    precision: jax.lax.PrecisionLike = None,
    preferred_element_type: jax.typing.DTypeLike | None = None,
    group_offset: jax.Array | None = None,
):
  """Quantized ragged_dot_general with a fast path."""
  if isinstance(lhs, qarray.QArray):
    lhs_val = lhs.qvalue
    lhs_scale = lhs.scale
    lhs_tiled_axes = qarray.get_tiled_axes(lhs)
  else:
    lhs_val = lhs
    lhs_scale = None
    lhs_tiled_axes = {}
  if isinstance(rhs, qarray.QArray):
    rhs_val = rhs.qvalue
    rhs_scale = rhs.scale
    rhs_tiled_axes = qarray.get_tiled_axes(rhs)
  else:
    rhs_val = rhs
    rhs_scale = None
    rhs_tiled_axes = {}

  (lhs_ca, rhs_ca), (lhs_ba, rhs_ba) = dimension_numbers.dot_dimension_numbers

  # Figure out the tiled axes to use for the dot_general. For greater
  # flexibility, we allow a non-tiled axis to be contracted with a tiled axis.
  # However, if both axes are tiled, their tile sizes must be the same.
  lhs_tiled_ca = {}
  rhs_tiled_ca = {}
  for l, r in zip(lhs_ca, rhs_ca):
    lhs_tile_size = lhs_tiled_axes.get(l)
    rhs_tile_size = rhs_tiled_axes.get(r)
    if lhs_tile_size and rhs_tile_size and lhs_tile_size != rhs_tile_size:
      raise ValueError(
          'Contracting axes must be tiled with the same tile size.'
          f' {lhs_tiled_axes=} {rhs_tiled_axes=} {dimension_numbers=}'
      )
    if lhs_tile_size or rhs_tile_size:
      lhs_tiled_ca[l] = lhs_tile_size or rhs_tile_size
      rhs_tiled_ca[r] = lhs_tile_size or rhs_tile_size

  # Split lhs/rhs_value for tiled axes.
  lhs_val = qarray.split_axis(lhs_val, lhs_tiled_ca)
  rhs_val = qarray.split_axis(rhs_val, rhs_tiled_ca)

  lhs_ca, lhs_ba, sum_axes = _apply_tiling(lhs_ca, lhs_ba, lhs_tiled_ca)
  rhs_ca, rhs_ba, _ = _apply_tiling(rhs_ca, rhs_ba, rhs_tiled_ca)
  dot_dimension_numbers = (lhs_ca, rhs_ca), (lhs_ba, rhs_ba)
  dimension_numbers = jax.lax.RaggedDotDimensionNumbers(
      dot_dimension_numbers=dot_dimension_numbers,
      lhs_ragged_dimensions=dimension_numbers.lhs_ragged_dimensions,
      rhs_group_dimensions=dimension_numbers.rhs_group_dimensions,
  )

  preferred_element_type, result_type = qarray.get_accumulator_and_result_type(
      lhs, rhs, preferred_element_type=preferred_element_type
  )

  out = jax.lax.ragged_dot_general(
      lhs_val,
      rhs_val,
      group_sizes,
      dimension_numbers,
      precision=precision,
      preferred_element_type=preferred_element_type,
      group_offset=group_offset,
  )

  lhs_scale_transpose, rhs_scale_transpose = _ragged_get_scale_transpose(
      dimension_numbers, (len(lhs_val.shape), len(rhs_val.shape))
  )
  if lhs_scale is not None:
    lhs_scale = qarray.split_axis(lhs_scale, {a: 1 for a in lhs_tiled_ca})
    lhs_scale = qarray.transpose_array(lhs_scale, lhs_scale_transpose)
    out = qarray.call_with_generic_broadcast(jnp.multiply, out, lhs_scale)
  if rhs_scale is not None:
    rhs_scale = qarray.split_axis(rhs_scale, {a: 1 for a in rhs_tiled_ca})
    # Check if the scale has a group dimension that needs special handling.
    if (
        dimension_numbers.rhs_group_dimensions
        and rhs_scale.shape[dimension_numbers.rhs_group_dimensions[0]] > 1
    ):
      rhs_scale = _apply_group_channelwise_scale(
          rhs_scale,
          lhs_val.shape,
          group_sizes,
          dimension_numbers,
          precision,
          group_offset,
      )
    else:
      rhs_scale = qarray.transpose_array(rhs_scale, rhs_scale_transpose)
    out = qarray.call_with_generic_broadcast(jnp.multiply, out, rhs_scale)

  if sum_axes:
    # [tile_count1, tile_count2, ..., M, N] -> [M, N]
    out = jnp.sum(out, axis=sum_axes)

  return out.astype(result_type)


def _slow_ragged_dot_general(
    lhs: qarray.MaybeQArray,
    rhs: qarray.MaybeQArray,
    group_sizes: jax.Array,
    dimension_numbers: jax.lax.RaggedDotDimensionNumbers,
    **kwargs,
):
  """A ragged_dot_general which dequantizes first."""
  lhs = qarray.dequantize(lhs) if isinstance(lhs, qarray.QArray) else lhs
  rhs = qarray.dequantize(rhs) if isinstance(rhs, qarray.QArray) else rhs
  return jax.lax.ragged_dot_general(
      lhs, rhs, group_sizes, dimension_numbers, **kwargs
  )


def ragged_dot_general(
    lhs: qarray.MaybeQArray,
    rhs: qarray.MaybeQArray,
    group_sizes: jax.Array,
    dimension_numbers: jax.lax.RaggedDotDimensionNumbers,
    precision: jax.lax.PrecisionLike = None,
    preferred_element_type: jax.typing.DTypeLike | None = None,
    group_offset: jax.Array | None = None,
) -> jax.Array:
  """Quantized jax.lax.ragged_dot_general."""
  use_fast_path = True
  for operand in (lhs, rhs):
    if isinstance(operand, qarray.QArray):
      if operand.zero_point is not None:
        use_fast_path = False
        break
    else:
      if numerics.should_quantize(operand.dtype):
        # Always dequantize on inputs if any of the operands is in bf16/fp32,
        # because XLA is able to fuse the dequantize and the matmul. The slow
        # path is usually not slower than the fast path, since both use fp
        # matmul, and will be significantly faster when subchannel or zero_point
        # is used.
        use_fast_path = False
        break

  if use_fast_path:
    return _fast_ragged_dot_general(
        lhs,
        rhs,
        group_sizes,
        dimension_numbers,
        precision=precision,
        preferred_element_type=preferred_element_type,
        group_offset=group_offset,
    )
  else:
    return _slow_ragged_dot_general(
        lhs,
        rhs,
        group_sizes,
        dimension_numbers,
        precision=precision,
        preferred_element_type=preferred_element_type,
        group_offset=group_offset,
    )


def ragged_dot(
    lhs: qarray.MaybeQArray,
    rhs: qarray.MaybeQArray,
    group_sizes: jax.Array,
    precision: jax.lax.PrecisionLike = None,
    preferred_element_type: jax.typing.DTypeLike | None = None,
    group_offset: jax.Array | None = None,
) -> jax.Array:
  """Quantized jax.lax.ragged_dot."""
  return ragged_dot_general(
      lhs,
      rhs,
      group_sizes,
      dimension_numbers=_BASIC_RAGGED_DOT_DIMENSION_NUMBERS,
      precision=precision,
      preferred_element_type=preferred_element_type,
      group_offset=group_offset,
  )



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utility functions to work with Flax linen and nnx modules."""

from collections.abc import Callable, Collection, Sequence
import dataclasses
import inspect
from typing import Any

import flax
from flax import linen as nn
from flax import nnx
import jax


@nnx.register_variable_name('quant_stats', overwrite=True)
class QuantStat(nnx.Variable):
  """NNX variable that stores quantization statistics."""


def should_update_quant_stats() -> bool:
  """Returns whether the current module should update the quantization statistics."""
  module = get_current_module()
  match module:
    case nn.Module():
      if module.is_initializing():
        return False  # Don't update quant_stats during initialization.
      return module.is_mutable_collection('quant_stats')
    case nnx.Module():
      return not module.disable_quant_stats_update


def get_current_module() -> nn.Module | nnx.Module:
  """Returns the current module, or raises an error if current module is not known.

  Args: None.

  Returns:
    The current module, it can be either a linen module or an nnx module.
  Raises:
    ValueError: If the current module is not known.
  """
  # Linen modules can be found in nn.module._context.module_stack.
  if nn.module._context.module_stack and nn.module._context.module_stack[-1]:  # pylint: disable=protected-access
    return nn.module._context.module_stack[-1]  # pylint: disable=protected-access

  # Retrieves the nnx modules from the caller stack.
  current_frame = inspect.currentframe()
  while current_frame is not None:
    if 'self' in current_frame.f_locals and isinstance(
        current_frame.f_locals['self'], nnx.Module
    ):
      return current_frame.f_locals['self']
    current_frame = current_frame.f_back

  raise ValueError('Current module is not known.')


def get_current_module_path() -> tuple[str, ...]:
  """Returns the path of the current module, or raises an error if current module is not known.

  Args: None.

  Returns:
    A tuple representing a unique path to the current module from the root
    module.
  """
  module = get_current_module()
  match module:
    case nn.Module():
      return module.path
    case nnx.Module():
      # Paths of nnx modules are set when they are quantized.
      return module.qwix_path


def get_or_create_variable(
    collection: str, name: str, init_fn: Callable[[], Any]
) -> nn.Variable | nnx.Variable:
  """Gets or creates a quant_stat in the current module.

  Args:
    collection: The variable collection name.
    name: The variable name.
    init_fn: The function that will be called to compute the initial value of
      the variable.

  Returns:
    A variable wrapped in nn.Variable or nnx.Variable.
  """
  module = get_current_module()
  if isinstance(module, nn.Module):
    # Instead of module.variable, use module.scope.variable which allows us to
    # create variables in non-compact modules.
    assert module.scope is not None
    variable = module.scope.variable(collection, name, init_fn)
    _check_shape(variable.value, init_fn)
    return variable
  elif isinstance(module, nnx.Module):
    var_type = nnx.variable_type_from_name(collection)
    if hasattr(module, name):
      variable = getattr(module, name)
      assert type(variable) is var_type, f'{type(variable)} != {var_type}'  # pylint: disable=unidiomatic-typecheck
      _check_shape(variable.value, init_fn)
    else:
      variable = var_type(init_fn())
      setattr(module, name, variable)
    return variable


def get_and_delete_variable(collection: str, name: str) -> Any | None:
  """Gets and deletes a quant_stat in the current module.

  This is mainly for NNX and doesn't make much sense for linen.

  Args:
    collection: The variable collection name, only used for linen modules.
    name: The variable name.

  Returns:
    The variable, or None if it doesn't exist.
  """
  module = get_current_module()
  if isinstance(module, nn.Module):
    assert module.scope is not None
    if module.has_variable(collection, name):
      return module.scope._mutable_collection(collection).pop(name)  # pylint: disable=protected-access
    return None
  elif isinstance(module, nnx.Module):
    quant_stat = getattr(module, name, None)
    if quant_stat is not None:
      delattr(module, name)
    return quant_stat


def get_or_create_param(
    name: str,
    init_fn: Callable[[], Any],
    nnx_param_type: type[nnx.Param] = nnx.Param,
) -> Any:
  """Gets or creates a param in the current module.

  Args:
    name: The param name.
    init_fn: The function that will be called to compute the initial value of
      this variable.
    nnx_param_type: The nnx param type of the param, if it's not a nnx.Param.

  Returns:
    The unboxed param.
  """
  module = get_current_module()
  if isinstance(module, nn.Module):
    # Instead of module.variable, use module.scope.variable which allows us to
    # create params in non-compact modules.
    assert module.scope is not None
    return module.scope.param(name, lambda _: init_fn())
  elif isinstance(module, nnx.Module):
    if hasattr(module, name):
      param = getattr(module, name)
      _check_shape(param.value, init_fn)
    else:
      param = jax.tree.map(nnx_param_type, init_fn())
      setattr(module, name, param)
    return unbox(param)


def find_param(x: Any) -> str | None:
  """Finds the param name of a given array in the current module.

  This function is useful when
    * We want to know if an array is a weight to decide how to quantize it.
    * We want to locate and replace the weights in PTQ mode.
    * We want to create lora weights together with the base weights.

  This function is designed so that array and array-like objects are treated
  equally.

  Args:
    x: jax.Array or array-like object (such as ptq.WithAux) that has the "shape"
      attribute.

  Returns:
    The name of the param that contains the given array, or None if not found.

  Raises:
    ValueError: If multiple params are found.
  """
  # First, identify all the candidates. We support the following cases:
  # * The type of the param is promoted, e.g., astype.
  # * The sharding constraints are added, i.e. with_sharding_constraint.
  # * The param is reshaped (only through the approach 1 below).
  module = get_current_module()
  candidates: dict[str, Any] = {}
  if isinstance(module, nn.Module):
    assert module.scope is not None
    for name, value in module.scope._collection('params').items():  # pylint: disable=protected-access
      value = nn.unbox(value)
      if hasattr(value, 'shape'):
        candidates[name] = value
  elif isinstance(module, nnx.Module):
    for name, node in module.__dict__.items():
      if hasattr(node, 'shape'):
        candidates[name] = node.value
  else:
    raise ValueError('Current module is not known.')

  candidates_by_id = {id(c): n for n, c in candidates.items()}

  # Approach 0: check if we could find the exact x in the params.
  if id(x) in candidates_by_id:
    return candidates_by_id[id(x)]

  # Approach 1: if x is a JitTracer, we can actually trace it back to the param.
  if isinstance(x, jax.core.Tracer) and hasattr(x, 'parent'):
    while True:
      if id(x) in candidates_by_id:
        return candidates_by_id[id(x)]
      if x.parent and len(x.parent.in_tracers) == 1:
        # Allow any unary primitives to be applied to the param, including
        # reshape, with_sharding_constraint, astype, etc.
        x = x.parent.in_tracers[0]
      elif id(const := x.get_const()) in candidates_by_id:
        # Even if x is a tracer, the candidate might be a concrete value.
        return candidates_by_id[id(const)]
      else:
        return None

  # Approach 2: use heuristics on the shape which may not be accurate.
  candidates = {n: c for n, c in candidates.items() if c.shape == x.shape}
  if len(candidates) > 2:
    raise ValueError(f'Multiple candidate params found: {candidates.keys()}')
  if len(candidates) == 1:
    return list(candidates.keys())[0]

  return None


def unbox(maybe_boxed: Any) -> Any:
  """Similar to nn.unbox but also works for nnx.Variable/VariableState/VariableMetadata."""

  def fn(x):
    if isinstance(x, nnx.Variable):
      return x.value
    elif isinstance(x, nnx.VariableState):
      return x.value
    elif isinstance(x, nnx.VariableMetadata):
      return x.raw_value
    elif isinstance(x, nn.meta.AxisMetadata):
      return x.unbox()
    else:
      return x

  return jax.tree.map(
      fn,
      maybe_boxed,
      is_leaf=lambda x: isinstance(x, nn.meta.AxisMetadata | nnx.Variable),
  )


def update_sharding(
    spec: Sequence[Any],
    *,
    shape: Sequence[int] | None = None,
    split: Collection[int] | None = None,
    merge: Collection[int] | None = None,
    transpose: Sequence[int | None] | None = None,
) -> tuple[Any, ...]:
  """Derives the partition spec from an existing spec.

  Args:
    spec: The existing partition spec.
    shape: Optional. The shape of the value. If provided, the partition spec
      will be updated to remove sharding for dimensions of size 1.
    split: Splits the given spec into (metadata, None).
    merge: Merge the given spec (metadata, None) into metadata. This is the
      inverse of split and takes the same arguments.
    transpose: Transposes the spec.

  Returns:
    The updated partition spec.
  """
  assert bool(split) + bool(merge) + bool(transpose) <= 1
  if split:
    spec = [(a, None) if i in split else (a,) for i, a in enumerate(spec)]
    spec = sum(spec, ())  # flatten the list of tuples.
  elif merge:
    for i in merge:
      spec = spec[: i + 1] + spec[i + 2 :]  # pytype: disable=unsupported-operands
  elif transpose:
    spec = tuple(spec[i] if i is not None else None for i in transpose)

  if shape:
    assert len(shape) == len(spec), f'{shape=} {spec=}'
    # For scales: remove sharding for dimensions of size 1.
    spec = tuple(None if d == 1 else a for a, d in zip(spec, shape))

  return spec


def update_boxed(
    boxed: nn.meta.AxisMetadata | nnx.Variable | jax.Array,
    *,
    value: jax.Array | None = None,
    split: Collection[int] | None = None,
    merge: Collection[int] | None = None,
    transpose: Sequence[int | None] | None = None,
) -> nn.meta.AxisMetadata | nnx.Variable | jax.Array:
  """Update a boxed value and its metadata, i.e., nn.meta.AxisMetadata or nnx.Variable.

  The most common use case of metadata is to store the logical axis names for
  sharding.

  This method can split, merge, or transpose the axes metadata. Only one of
  them can be used at a time.

  Args:
    boxed: The metadata to update, or a jax.Array if no metadata is available.
    value: The value used to replace the boxed value.
    split: Splits the given axes metadata into (metadata, None).
    merge: Merge the given axes (metadata, None) into metadata. This is the
      inverse of split and takes the same arguments.
    transpose: Transposes the axes metadata.

  Returns:
    The updated metadata, which may be the same as the input.
  """
  if isinstance(boxed, nn.meta.AxisMetadata):
    if value is not None:
      boxed = boxed.replace_boxed(value)
    shape = boxed.unbox().shape
    for possible_field in ('names', 'mesh_axes', 'axes_types'):
      axes = getattr(boxed, possible_field, None)
      if isinstance(axes, (list, tuple)):
        axes = update_sharding(
            axes, shape=shape, split=split, merge=merge, transpose=transpose
        )
        boxed = dataclasses.replace(boxed, **{possible_field: axes})
  elif isinstance(boxed, nnx.Variable):
    if value is not None:
      boxed = boxed.replace(value)
    shape = boxed.value.shape
    axes = boxed.get_metadata().get('sharding_names', None)
    if isinstance(axes, (list, tuple)):
      axes = update_sharding(
          axes, shape=shape, split=split, merge=merge, transpose=transpose
      )
      boxed.set_metadata(sharding_names=axes)
  elif isinstance(boxed, jax.Array):  # not boxed.
    if value is not None:
      boxed = value
  else:
    raise ValueError(f'Unsupported type: {type(boxed)}')

  return boxed


def _check_shape(value: Any, init_fn: Callable[[], Any]):
  """Checks the shape of the value to match the shape of init_fn().

  This is similar to the check in nn.Module.param, which is not performed for
  linen variables or in nnx.

  Args:
    value: The value to check.
    init_fn: The function that will be called to compute the initial value of
      this variable.

  Raises:
    ValueError: If the shape of the value does not match the shape of init_fn().
  """
  # Value may be a FrozenDict.
  abs_value = jax.eval_shape(lambda: flax.core.unfreeze(value))
  abs_init_value = jax.eval_shape(lambda: unbox(init_fn()))
  if abs_value != abs_init_value:
    raise ValueError(f'{abs_value} != {abs_init_value}')


def make_rng(rng_stream: str) -> jax.Array:
  """Generate a random key from the given rng_stream in the current module."""
  module = get_current_module()
  if isinstance(module, nn.Module):
    return module.make_rng(rng_stream)
  elif isinstance(module, nnx.Module):
    rngs = getattr(module, 'rngs', None)
    if not isinstance(rngs, nnx.Rngs):
      raise ValueError('Cannot find rngs in the current module.')
    return rngs[rng_stream]()
  else:
    raise ValueError('Current module is not known.')



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Qwix APIs for quantizing Pallas kernels.

This module is usually imported as `import qwix.pallas as qpl` in user code.

To add quantization support to a Pallas kernel, generally users need to

1) Quantize the inputs outside of the kernel via `qpl.quantize`.
2) Instead of calling `pl.pallas_call`, call `qpl.pallas_call`, which supports
   taking QArrays as inputs. `qpl.pallas_call` takes the same block specs as
   `pl.pallas_call` and adjusts them automatically for QArrays.
3) When inside the kernel, use `qpl.einsum` / `qpl.dot_general` / `qpl.dot`
   which take QArrays as inputs and return jax.Array. Alternatively, call
   `dequantize` on the QArray to obtain the dequantized jax.Array.

Note: dot_general / einsum / dot in this module should only be called inside
pallas kernels, as they can be inefficient when called outside of the kernel.
"""

import functools

from qwix._src import qconfig
from qwix._src.core import dot as qwix_dot
from qwix._src.core import dot_general as qwix_dot_general
from qwix._src.core import einsum as qwix_einsum
from qwix._src.core import pallas
from qwix._src.core import qarray

__all__ = [
    'QArray',
    'quantize',
    'dequantize',
    'get_current_rule',
    'pallas_call',
    'dot_general',
    'dot',
    'einsum',
]

# APIs for generic QArray operations. These are deprecated in favor of the
# qwix.* APIs.
QArray = qarray.QArray
quantize = qarray.quantize_api
dequantize = qarray.dequantize

# APIs for quantizing Pallas kernels.
get_current_rule = qconfig.get_current_rule
pallas_call = pallas.pallas_call

# Use loop-based implementations for Pallas kernels.
dot_general = qwix_dot_general.loop_dot_general
dot = functools.partial(qwix_dot.dot, _qwix_dot_general=dot_general)
einsum = functools.partial(qwix_einsum.einsum, _qwix_dot_general=dot_general)



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Averaging of calibration into quant stats."""

from typing import TypeAlias

import jax
import jax.numpy as jnp

QuantStat: TypeAlias = dict[str, jax.Array]
Calibration: TypeAlias = dict[str, jax.Array]


class SimpleMovingAverage:
  """Simple moving average maintains a count and sums of the calibration."""

  def __init__(self, bootstrap_steps: int = 0):
    """Initializes the simple moving average.

    Args:
      bootstrap_steps: The number of steps to bootstrap the quantization
        statistics.
    """
    self._bootstrap_steps = bootstrap_steps

  def init(self, calibration: Calibration) -> QuantStat:
    """Initializes the quantization statistics from the calibration."""
    quant_stat = {'count': jnp.zeros((), jnp.int32)}
    for key, value in calibration.items():
      # Always use float32 for the sum, because bf16 only has 7 bits of
      # precision, which will cause accumulation becomes a no-op after a few
      # hundreds of steps.
      quant_stat[f'sum_of_{key}'] = jnp.zeros_like(value, dtype=jnp.float32)
    return quant_stat

  def update(
      self, quant_stat: QuantStat, calibration: Calibration
  ) -> QuantStat:
    """Updates the quantization statistics."""
    quant_stat = quant_stat.copy()
    quant_stat['count'] += 1
    for key, value in calibration.items():
      # Disallow automatic shape broadcasting which usually indicates a config
      # error.
      if quant_stat[f'sum_of_{key}'].shape != value.shape:
        raise ValueError(
            f'Shape mismatch between {quant_stat} and {calibration}'
        )
      quant_stat[f'sum_of_{key}'] += value
    return quant_stat

  def get_calibration(
      self,
      quant_stat: QuantStat,
      default_calibration: Calibration | None = None,
  ) -> Calibration:
    """Returns the average calibration.

    If a default calibration is provided, it will be used when the number of
    samples is less than the bootstrap steps.

    Args:
      quant_stat: The quantization statistics.
      default_calibration: The default calibration to use when the number of
        samples is less than the bootstrap steps.

    Returns:
      The average calibration.
    """
    calibration = {}
    for key, value in quant_stat.items():
      if key.startswith('sum_of_'):
        calibration[key.removeprefix('sum_of_')] = value / quant_stat['count']

    if default_calibration is not None:
      calibration = jax.tree.map(
          lambda x, y: x.astype(y.dtype), calibration, default_calibration
      )
      return jax.lax.cond(
          quant_stat['count'] > self._bootstrap_steps,
          lambda: calibration,
          lambda: default_calibration,
      )

    return calibration
